/*******************************************************************************
  File: BtreeIndex_Page.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
    The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "BtreeIndex.h"


namespace Util {

// XDR description
//  Page {
//      unsigned int    num_keys;
//      unsigned int    prev_page;
//      unsigned int    next_page;
//      unsigned int    dummy;  // always 0
//      opaque          data[PageSize - 16]; -- PageSize minus the header variables
//  };


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex_Page<Key,Compare,KeyTraits>::BtreeIndex_Page()
  : _pageId(0),
    _bufSize(0),
    _dataLen(0),
    _prevPage(-1),
    _nextPage(-1),
    _dirty(false),
    _keyList()
{
    // empty
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex_Page<Key,Compare,KeyTraits>::BtreeIndex_Page(size_t pid, size_t pgSize)
  : _pageId(pid),
    _bufSize( (pgSize - HeaderSize) / sizeof(XDR_Char) ),
    _dataLen(0),
    _prevPage(-1),
    _nextPage(-1),
    _dirty(false),
    _keyList()
{
    // empty
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex_Page<Key,Compare,KeyTraits>::~BtreeIndex_Page()
{
    // empty
}


template<typename Key, typename Compare, typename KeyTraits>
bool
BtreeIndex_Page<Key,Compare,KeyTraits>::must_split() const
{
    return _dataLen > _bufSize;
}


template<typename Key, typename Compare, typename KeyTraits>
std::pair<size_t,bool>
BtreeIndex_Page<Key,Compare,KeyTraits>::add_key(BtreeIndex& btree, 
                                                const std::pair<Key, size_t>& x)
{
    size_t slot = 0;
    // Find where the key should go
    KeyList::iterator kit = _keyList.begin();
    for (; kit != _keyList.end(); ++kit) {
        if (btree._uniqueKeys) {
            if (not btree._comp((*kit).key, x.first)) break;    // lower bound
        } else {
            if (btree._comp(x.first, (*kit).key)) break;        // upper bound
        }
        ++slot;
    }
    if (slot < _keyList.size()      // do we already have this key
            and btree._uniqueKeys
                and not btree._comp(x.first, (*kit).key)) 
        return std::make_pair(slot,false);


    // Encode the <key,link> to see how long it is
    XDR_Char* buffer = new XDR_Char[_bufSize];
    XOutDataBuf xstrm(btree._xstream, buffer, _bufSize);

    KeyTraits::encode(xstrm, x.first) << x.second;
    size_t keyLen = xstrm.tellp();
    if (keyLen > _bufSize / 2) 
        throw std::ios_base::failure("BtreeIndex::insert: key too large");
    
    KeyInfo info(x.first, x.second, keyLen);
    kit = _keyList.insert(kit, info);
    _dataLen += info.xlen;
    _dirty = true;
    delete[] buffer;
    return std::make_pair(slot,true);
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex_Page<Key,Compare,KeyTraits>::remove_key(size_t slot)
{
    if (slot >= _keyList.size())
        throw std::ios_base::failure("BtreeIndex::erase: invalid iterator");

    // Find the slot
    KeyList::iterator kit = _keyList.begin();
    for (int ii = 0; size_t(ii) < slot; ++ii)
        ++kit;
    _dataLen -= (*kit).xlen;
    _keyList.erase(kit);
    _dirty = true;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex_Page<Key,Compare,KeyTraits>::update_key(size_t slot, size_t link)
{
    // Find the slot
    KeyList::iterator kit = _keyList.begin();
    for (int ii = 0; size_t(ii) < slot; ++ii)
        ++kit;
    (*kit).link = link;
    _dirty = true;
    return;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex_Page<Key,Compare,KeyTraits>::insert_page_link(KeyList::iterator kit, Page* p)
{
    // Build a KeyInfo object from the link page
    KeyInfo info = p->key_list().front();
    info.link = p->page_id();   // doesn't change xlen
    _dataLen += info.xlen;
    _keyList.insert(kit, info);
    _dirty = true;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex_Page<Key,Compare,KeyTraits>::update_page_link(Page* p)
{
    // The first key on the decendant page has changed
    // Find the page reference
    KeyList::iterator kit = _keyList.begin();
    for (; kit != _keyList.end(); ++kit) {
        if ((*kit).link == p->page_id()) break;
    }

    // Carefully replace the keyinfo at this slot
    _dataLen -= (*kit).xlen;
    (*kit) = p->key_list().front();
    (*kit).link = p->page_id();
    _dataLen += (*kit).xlen;
    _dirty = true;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex_Page<Key,Compare,KeyTraits>::remove_page_link(Page* p)
{
    // Find the page reference
    KeyList::iterator kit = _keyList.begin();
    for (; kit != _keyList.end(); ++kit) {
        if ((*kit).link == p->page_id()) break;
    }
    if (kit == _keyList.end())
        throw std::ios_base::failure("BtreeIndex::Page::remove_page_link: invalid page reference");

    // Remove it
    _dataLen -= (*kit).xlen;
    _keyList.erase(kit);
    _dirty = true;
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex_Page<Key,Compare,KeyTraits>*
BtreeIndex_Page<Key,Compare,KeyTraits>::split(BtreeIndex& btree, Iterator& it)
{
    // Go through the page and find where to split it
    KeyList::iterator kit = _keyList.begin();
    size_t len = 0;
    while (len < _bufSize / 2) {
        len += (*kit).xlen;
        ++kit;
    }
    
    // Now move the rest of the elements to a new page
    Page* newp = btree.get_free_page();
    KeyList& kl = newp->_keyList;
    kl.splice(kl.end(), _keyList, kit, _keyList.end());
    newp->_dataLen = _dataLen - len;
    _dataLen = len;

    // link the page in after this one
    newp->next_page(_nextPage);
    newp->prev_page(_pageId);
    Page* nextp = btree.get_page(_nextPage);
    if (nextp) nextp->prev_page(newp->page_id());
    _nextPage = newp->page_id();

    // make sure both pages get flushed
    _dirty = newp->_dirty = true;

    // Update the return iterator if necessary
    if (it._page == _pageId and it._slot >= num_keys()) {
        it._page = newp->_pageId;
        it._slot -= num_keys();
    }

    // Return the new page
    return newp;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex_Page<Key,Compare,KeyTraits>::compact(BtreeIndex& btree, Iterator& pos)
{
    if (_dataLen >= _bufSize / 2) return;

    // Compaction can only be done with the previous page
    Page* prevp = btree.get_page(_prevPage);
    if (prevp and (prevp->data_len() + _dataLen) < _bufSize) {
        size_t tmp = num_keys();
        KeyList& kl = prevp->key_list();
        kl.splice(kl.end(), _keyList);
        prevp->_dataLen += _dataLen;
        _dataLen = 0;
        if (pos._page == _pageId) {
            pos._page = prevp->page_id();
            pos._slot += prevp->num_keys() - tmp;
        }
        _dirty = prevp->_dirty = true;
    } else if (prevp) {
        KeyList& kl = prevp->key_list();
        _keyList.splice(_keyList.begin(), kl, --kl.end());
        prevp->_dataLen -= _keyList.front().xlen;
        _dataLen += _keyList.front().xlen;
        if (pos._page == _pageId) {
            pos._slot += 1;
        }
        _dirty = prevp->_dirty = true;
    }
}


template<typename Key, typename Compare, typename KeyTraits>
oXDR_Stream&
BtreeIndex_Page<Key,Compare,KeyTraits>::encode(oXDR_Stream& xstrm)
{
    xstrm.seekp(_pageId);

    xstrm << _keyList.size();
    xstrm << _prevPage;
    xstrm << _nextPage;
    xstrm << 0;

    XDR_Char* buffer = new XDR_Char[_bufSize];
    XOutDataBuf xdatastrm(xstrm, buffer, _bufSize);

    KeyList::iterator it = _keyList.begin(); 
    for (; it != _keyList.end(); ++it) {
        KeyTraits::encode(xdatastrm, (*it).key) << (*it).link;
    }
    xstrm.put_opaque(buffer, _bufSize * sizeof(XDR_Char));

    _dirty = false;
    delete[] buffer;
    return xstrm;
}


template<typename Key, typename Compare, typename KeyTraits>
iXDR_Stream&
BtreeIndex_Page<Key,Compare,KeyTraits>::decode(iXDR_Stream& xstrm)
{
    xstrm.seekg(_pageId);

    unsigned long numKeys;
    xstrm >> numKeys;
    xstrm >> _prevPage;
    xstrm >> _nextPage;
    unsigned long dummy;
    xstrm >> dummy;

    XDR_Char* buffer = new XDR_Char[_bufSize];
    xstrm.get_opaque(buffer, _bufSize * sizeof(XDR_Char));

    XInDataBuf xdatastrm(xstrm, buffer, _bufSize);

    _dataLen = 0;
    _keyList.clear();
    while (numKeys-- > 0) {
        KeyInfo info;
        info.xlen = short(xdatastrm.tellg()); // the start
        KeyTraits::decode(xdatastrm, info.key);
        xdatastrm >> info.link;
        info.xlen = short(xdatastrm.tellg()) - info.xlen;
        _dataLen += info.xlen;
        _keyList.push_back(info);
    }
    delete[] buffer;
    return xstrm;
}


} //> namespace
