/*******************************************************************************
  File: BtreeIndex_RootPage.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
    The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "BtreeIndex.h"


namespace Util {

// XDR description
//  RootPage {
//      unsigned int    unique + page_size + depth; -- 2 short ints plus a bool
//      unsigned int    total_keys;
//      unsigned int    num_keys;
//      unsigned int    free_page;
//      opaque          data[PageSize - 16]; -- PageSize minus the header variables
//  };


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex_RootPage<Key,Compare,KeyTraits>::BtreeIndex_RootPage(iXDR_Stream& xstrm)
  : Page(),
    _pageSize(0),
    _totalKeys(0),
    _depth(0),
    _dirtyTotal(false)
{
    decode(xstrm);
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex_RootPage<Key,Compare,KeyTraits>::BtreeIndex_RootPage(
        size_t pageSize, bool uniqueKeys)
  : Page(0, pageSize),
    _pageSize(pageSize),
    _uniqueKeys(uniqueKeys),
    _totalKeys(0),
    _depth(0),
    _dirtyTotal(false)
{
    // empty
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex_RootPage<Key,Compare,KeyTraits>::~BtreeIndex_RootPage()
{
    // empty
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex_Page<Key,Compare,KeyTraits>*
BtreeIndex_RootPage<Key,Compare,KeyTraits>::split(BtreeIndex& btree, Iterator& it)
{
    // Find where to split the page
    KeyList::iterator kit = _keyList.begin();
    size_t len = 0;
    while (len < _bufSize / 2) {
        len += (*kit).xlen;
        ++kit;
    }

    // Spliting the root page means creating two new pages
    Page* newp1 = btree.get_free_page();
    Page* newp2 = btree.get_free_page();

    // move the first part to new page 1
    KeyList& kl1 = newp1->key_list();
    kl1.splice(kl1.end(), _keyList, _keyList.begin(), kit);
    newp1->data_len(len);

    // move whats left to the second new page
    KeyList& kl2 = newp2->key_list();
    kl2.splice(kl2.end(), _keyList);
    newp2->data_len(_dataLen - len);
    _dataLen = 0;

    // link the two pages together
    newp1->next_page(newp2->page_id());
    newp2->prev_page(newp1->page_id());

    // put the page links into the now empty root page
    insert_page_link(_keyList.end(), newp1);
    insert_page_link(_keyList.end(), newp2);
    _depth += 1;
    _dirty = true;
    btree._pageCache.min_size(_depth);

    // Update the iterator if necessary
    if (it._page == _pageId) {
        if (it._slot < newp1->num_keys()) {
            it._page = newp1->page_id();
        } else {
            it._page = newp2->page_id();
            it._slot -= newp1->num_keys();
        }
    }
    return 0;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex_RootPage<Key,Compare,KeyTraits>::compact(BtreeIndex& btree, Iterator& pos)
{
    if (depth() == 0 or num_keys() > 1) return;

    // consolodating the root page means eliminating the lower level page
    Page* oldp = btree.get_page(_keyList.front().link);
    _keyList.clear();

    // move the keys from the lower level page
    _keyList.splice(_keyList.begin(), oldp->key_list());
    _dataLen = oldp->data_len();
    oldp->data_len(0);
    if (pos._page == oldp->page_id()) 
        pos._page = 0;
    
    btree.add_free_page(oldp);
    _depth -= 1;
    _dirty = true;
    btree._pageCache.min_size(_depth);
}


template<typename Key, typename Compare, typename KeyTraits>
oXDR_Stream&
BtreeIndex_RootPage<Key,Compare,KeyTraits>::encode(oXDR_Stream& xstrm)
{
    xstrm.seekp(_pageId);

    unsigned long first = (_uniqueKeys) ? 0x80000000 : 0x00000000;
    first |= page_size() << 16;
    first |= _depth;

    xstrm << first;
    xstrm << _totalKeys;
    xstrm << _keyList.size();   // num_keys;
    xstrm << _nextPage;         // free page id

    XDR_Char* buffer = new XDR_Char[_bufSize];
    XOutDataBuf xdatastrm(xstrm, buffer, _bufSize);

    KeyList::iterator it = _keyList.begin(); 
    for (; it != _keyList.end(); ++it) {
        KeyTraits::encode(xdatastrm, (*it).key) << (*it).link;
    }
    xstrm.put_opaque(buffer, _bufSize * sizeof(XDR_Char));
    delete[] buffer;
    _dirty = _dirtyTotal = false;
    return xstrm;
}


template<typename Key, typename Compare, typename KeyTraits>
iXDR_Stream&
BtreeIndex_RootPage<Key,Compare,KeyTraits>::decode(iXDR_Stream& xstrm)
{
    xstrm.seekg(_pageId);

    unsigned long first;
    xstrm >> first;
    _pageSize = (first bitand 0x7fff0000) >> 16;
    _bufSize = (_pageSize - HeaderSize) / sizeof(XDR_Char);

    _uniqueKeys = (first bitand 0x80000000) != 0;

    _depth = short(first bitand 0x0000ffff);
    xstrm >> _totalKeys;

    unsigned long numKeys;
    xstrm >> numKeys;

    xstrm >> _nextPage;

    XDR_Char* buffer = new XDR_Char[_bufSize];
    xstrm.get_opaque(buffer, _bufSize * sizeof(XDR_Char));
    XInDataBuf xdatastrm(xstrm, buffer, _bufSize);

    _dataLen = 0;
    _keyList.clear();
    while (numKeys-- > 0) {
        KeyInfo info;
        info.xlen = short(xdatastrm.tellg());
        KeyTraits::decode(xdatastrm, info.key);
        xdatastrm >> info.link;
        info.xlen = short(xdatastrm.tellg()) - info.xlen;
        _dataLen += info.xlen;
        _keyList.push_back(info);
    }
    delete[] buffer;
    return xstrm;
}


} //> namespace
