/*******************************************************************************
  File: IndexedFile.h
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/
#ifndef UTIL_INDEXED_FILE_H
#define UTIL_INDEXED_FILE_H

// Platform header
#include "platform.h"

// Index header
#include "BtreeIndex.h"

// Utility headers
#include "check.h"

// XDR Stream headers
#include "XDRStream.h"
#include "XIO.h"


// Standard C++ headers
#include <functional>
#include <string>
#include <ios>
#include <list>

namespace Util {


// Forward declarations
template<typename Key,
         typename Item,
         typename Compare = std::less<Key>,
         typename KeyTraits = XDR_Traits<Key>,
         typename ItemTraits = XDR_Traits<Item> >
class IndexedFile;


//
// IndexedFile_Iterator --
//
template<typename Key,
         typename Item,
         typename Compare,
         typename KeyTraits,
         typename ItemTraits>
class IndexedFile_Iterator : public std::iterator<
                                    std::bidirectional_iterator_tag,
                                    std::pair<Key,Item> > {
    // friends
    friend class IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>;
    // types
    typedef IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>           IndexedFile;
    typedef IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>  Iterator;
    typedef BtreeIndex<Key,Compare,KeyTraits>                            index_type;
    typedef std::pair<Key,Item>                                          value_type;

    // data
    IndexedFile*                _ifile;
    index_type::iterator        _base;

public:
    IndexedFile_Iterator()  {}

    Iterator&           operator++();
    Iterator            operator++(int);
    Iterator&           operator--();
    Iterator            operator--(int);
    index_type::iterator base();

    bool operator==(const Iterator& x) const;
    bool operator!=(const Iterator& x) const;

    void                get_value(value_type& value);
    void                update_item(Item& item);

private:
    IndexedFile_Iterator(IndexedFile* ifile, index_type::iterator base);
};


//
// Class IndexedFile
//
template<typename Key,
         typename Item,
         typename Compare,
         typename KeyTraits,
         typename ItemTraits>
class IndexedFile {
public:
    // types
    typedef IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>  Iterator;

    typedef Key                         key_type;
    typedef std::pair<Key, Item>        value_type;
    typedef Compare                     key_compare;
    typedef Iterator                    iterator;
    typedef size_t                      size_type;
    typedef ptrdiff_t                   difference_type;
    typedef BtreeIndex<Key,Compare,KeyTraits> index_type;

    // construct / destroy
    explicit IndexedFile(Compare comp = Compare());
    explicit IndexedFile(const std::string& name, 
                            bool readOnly = false, 
                            Compare comp = Compare());
    ~IndexedFile();

    // open/close
    bool            is_open();
    void            create(const std::string& name, bool uniqueKeys = true);
    void            create(const std::string& name, int pageSize, bool uniqueKeys = true);
    void            open(bool readOnly = false);
    void            open(const std::string& name, bool readOnly = false);
    void            close();
    void            flush();

    // status check
    operator void*() const;
    bool operator!() const;

    // iterators
    Iterator        begin();
    Iterator        end();

    // capacity
    bool            empty();
    size_type       size();

    // modifiers
    std::pair<iterator,bool>
                    insert(const value_type& x);
    template<typename InputIterator>
        void        insert(InputIterator first, InputIterator last);

    iterator        erase(iterator pos);
    size_type       erase(const Key& x);
    iterator        erase(iterator first, iterator last);
    void            clear();

    // observers
    Compare         key_comp() const;
    index_type&     index();
    XDR_Stream&     xdrstream();

    // operations
    iterator        find(const Key& x);
    size_type       count(const Key& x);

    iterator        lower_bound(const Key& x);
    iterator        upper_bound(const Key& x);
    std::pair<iterator, iterator>
                    equal_range(const Key& x);

    // support for reference semantics and polymorphism
    Util::oXDR_Stream::ObjToIdMap&  obj2id();
    Util::oXDR_Stream::TypeToIdMap& type2id();
    Util::iXDR_Stream::IdToObjMap&  id2obj();
    Util::iXDR_Stream::IdToTypeMap& id2type();

    // utilities
    void            read(size_t pos, Item& item);  // decodes and returns the Item at pos

private:
    // data
    std::string             _name;
    index_type              _index;
    ItemTraits::xio_type    _xstream;

    // friendly utilities
    size_t          write(const Item& item); // returns the streampos of the written item
};


//
// IndexedFile_Iterator inline function definitions
//
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>&
IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::operator++()
{ ++_base; return *this; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>
IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::operator++(int)
{ Iterator t = *this; ++_base; return t; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>&
IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::operator--()
{ --_base; return *this; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>
IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::operator--(int)
{ Iterator t = *this; --_base; return t; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::index_type::iterator
IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::base()
{ return _base; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline bool
IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::operator==(const Iterator& x) const
{ return _ifile == x._ifile and _base == x._base; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline bool
IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::operator!=(const Iterator& x) const
{ return !(*this == x); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline
IndexedFile_Iterator<Key,Item,Compare,KeyTraits,ItemTraits>::IndexedFile_Iterator(
            IndexedFile* ifile, 
            index_type::iterator base)
    : _ifile(ifile), _base(base) {}


//
// IndexedFile inline function definitions
//
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline bool
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::is_open()
{ return _index.is_open() and _xstream.is_open(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline 
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::operator void*() const
{ 
    if (!_index or !_xstream) return null;
    return const_cast<IndexedFile*>(this);
}

template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline bool
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::operator!() const
{ return (!_index or !_xstream); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::begin()
{ return Iterator(this, _index.begin()); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::end()
{ return Iterator(this, _index.end()); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline bool
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::empty()
{ return _index.empty(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::size_type
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::size()
{ return _index.size(); }


// Erase functions just erase the keys, the data becomes dead
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::erase(iterator pos)
{ return Iterator(this, _index.erase(pos.base())); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::size_type
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::erase(const Key& k)
{ return _index.erase(k); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::erase(Iterator first, Iterator last)
{ return Iterator(this, _index.erase(first.base(), last.base())); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::clear()
{ _index.clear(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline Compare
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::key_comp() const
{ return _index.key_comp(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline BtreeIndex<Key,Compare,KeyTraits>&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::index()
{ return _index; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline XDR_Stream&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::xdrstream()
{ return _xstream; }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::find(const Key& k)
{ return Iterator(this, _index.find(k)); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::size_type
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::count(const Key& k)
{ return _index.count(k); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::lower_bound(const Key& k)
{ return Iterator(this, _index.lower_bound(k)); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::upper_bound(const Key& k)
{ return Iterator(this, _index.upper_bound(k)); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline std::pair<IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator,
                 IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::equal_range(const Key& k)
{ return std::make_pair(lower_bound(k), upper_bound(k)); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline Util::oXDR_Stream::ObjToIdMap&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::obj2id()
{ return xdrstream().obj2id(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline Util::oXDR_Stream::TypeToIdMap&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::type2id()
{ return xdrstream().type2id(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline Util::iXDR_Stream::IdToObjMap&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::id2obj()
{ return xdrstream().id2obj(); }


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
inline Util::iXDR_Stream::IdToTypeMap&
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::id2type()
{ return xdrstream().id2type(); }


} //> namespace Util


#ifdef EXPORT_NOT_SUPPORTED
#include "IndexedFile.hxx"
#endif


#endif

