/*******************************************************************************
  File: BtreeIndex.h
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/
#ifndef UTIL_BTREE_INDEX_H
#define UTIL_BTREE_INDEX_H

// Platform header
#include "platform.h"

// XDR Stream headers
#include "XDRStream.h"
#include "XIO.h"


// Standard C++ headers
#include <functional>
#include <string>
#include <ios>
#include <list>

namespace Util {

//
// XDR_Traits --
//
template <typename Item>
struct XDR_Traits {
    typedef     XIOFileStream           xio_type;
    static oXDR_Stream& encode(oXDR_Stream& oxs, const Item& val)
        { return oxs << val; }
    static iXDR_Stream& decode(iXDR_Stream& ixs, Item& val)
        { return ixs >> val; }
};


//
// BtreeIndex --
//
template<typename Key,
         typename Compare = std::less<Key>,
         typename KeyTraits = XDR_Traits<Key> >
class BtreeIndex {
    // private types
    class XIODataStreambuf;
    class XInDataBuf;
    class XOutDataBuf;
    class Page;
    class RootPage;

public:
    // public types
    class Iterator;
    class StreamposProxy;

    typedef Key                         key_type;
    typedef std::pair<Key, size_t>      value_type;
    typedef Compare                     key_compare;
    typedef Iterator                    iterator;
    typedef size_t                      size_type;
    typedef ptrdiff_t                   difference_type;

    enum DeleteStrategy { None, Compact };
    enum UpdateStrategy { Unsafe, MostlySafe, Safe };

    class Iterator : public std::iterator<
                            std::bidirectional_iterator_tag, 
                            std::pair<Key, size_t> > {
        // friends
        friend class BtreeIndex;
        friend class Page;
        friend class RootPage;
        
        // types
        typedef std::pair<Key,size_t>       value_type;
        
        // data
        BtreeIndex*         _btree;
        size_t              _page;
        size_t              _slot;
    public:
        Iterator()  {}

        Iterator&           operator++();
        Iterator            operator++(int);
        Iterator&           operator--();
        Iterator            operator--(int);

        bool operator==(const Iterator& rhs) const;
        bool operator!=(const Iterator& rhs) const;

        value_type          get_value();
        void                update_link(size_t link);

    private:
        Iterator(BtreeIndex* btree, size_t page, size_t slot);
    };


    // construct / destroy
    explicit BtreeIndex(Compare comp = Compare());
    explicit BtreeIndex(const std::string& name, 
                            bool readOnly = false,
                            Compare comp = Compare());
    ~BtreeIndex();

    // open/close/test state
    bool            is_open();
    void            create(const std::string& name, bool uniqueKeys = true);
    void            create(const std::string& name, int pageSize, bool uniqueKeys = true);
    void            open(bool readOnly = false);
    void            open(const std::string& name, bool readOnly = false);
    void            close();
    void            flush();
    operator void*() const;
    bool operator!() const;

    // iterators
    iterator        begin();
    iterator        end();

    // capacity
    bool            empty();
    size_type       size();

    // modifiers
    std::pair<iterator,bool>
                    insert(const value_type& x);
    std::pair<iterator,bool>
                    insert(const Key& k);
    template<typename InputIterator>
        void        insert(InputIterator first, InputIterator last);

    iterator        erase(iterator pos);
    size_type       erase(const Key& x);
    iterator        erase(iterator first, iterator last);
    void            clear();

    // observers
    key_compare     key_comp() const;
    std::string     file_name() const;
    size_type       page_size() const;
    bool            unique_keys() const;
    bool            read_only() const;

    XDR_Stream&     xdrstream();

    // operations
    iterator        find(const Key& x);
    size_type       count(const Key& x);

    iterator        lower_bound(const Key& x);
    iterator        upper_bound(const Key& x);
    std::pair<iterator, iterator>
                    equal_range(const Key& x);

    // utilities
    void            reindex();
    DeleteStrategy  set_delete_strategy(DeleteStrategy strategy);
    UpdateStrategy  set_update_strategy(UpdateStrategy strategy);
    int             set_cache_size(int size);

private:
    class PageCache {
        // types
        typedef std::list<Page*> PageList;

        // data
        XDR_Stream& _xstream;
        size_t      _minSize;
        size_t      _maxSize;
        PageList    _pageList;
    public:
        PageCache(XDR_Stream&);
        ~PageCache();

        Page*       get_page(size_t pid, size_t pgSize);
        void        add_page(Page* p); // puts a new page into the cache
        void        reset();    // deletes any excess pages
        void        clear();    // deletes all pages
        void        flush();    // writes any dirty pages out

        void        min_size(size_t size)   { _minSize = size; }
        int         max_size()              { return _maxSize; }
        void        max_size(size_t size)   { _maxSize = size; }
    }; //> PageCache

    // friends
    friend class Iterator;
    friend class StreamposProxy;
    friend class Page;
    friend class RootPage;

    // data
    Compare                 _comp;
    std::string             _name;
    size_t                  _pageSize;
    bool                    _uniqueKeys;
    bool                    _readOnly;
    typename KeyTraits::xio_type _xstream;
    RootPage*               _root;
    PageCache               _pageCache;
    DeleteStrategy          _dStrategy;
    UpdateStrategy          _uStrategy;

    // Iterator utility functions - used by friends
    void            prev_iter(Iterator& it);
    void            next_iter(Iterator& it);
    void            get_value(Iterator& it, value_type& value);
    void            update_link(Iterator& it, size_t link);

    // BtreeIndex utility functions
    Page*           get_page(size_t pageId);
    Page*           get_free_page();
    void            add_free_page(Page* p);
    iterator        lower_bound(Page* p, int depth, const Key& k);
    iterator        upper_bound(Page* p, int depth, const Key& k);
    std::pair<iterator,bool>
                    insert(Page*, int depth, const value_type& x);
    bool            erase(Page* p, int depth, const Key& k, iterator& pos);

    void            flush(UpdateStrategy s);
}; //> BtreeIndex


//
// BtreeIndex::Page
//
template<typename Key, typename Compare, typename KeyTraits>
class BtreeIndex<Key,Compare,KeyTraits>::Page {
public:
    struct KeyInfo {
        Key         key;
        size_t      link;
        short       xlen;   // length of encoded <key,link> in data area
        KeyInfo(Key k, size_t l, short xl) : key(k), link(l), xlen(xl) {}
        KeyInfo() {}
    };
    typedef std::list<KeyInfo>  KeyList;

protected:
    // data
    const size_t _pageId;
    size_t      _bufSize;
    size_t      _dataLen;
    size_t      _prevPage;
    size_t      _nextPage;
    bool        _dirty;     // set when changed; cleared by encode
    KeyList     _keyList;

    enum { HeaderSize = 16 };

    Page();                 // used by RootPage

public:
    Page(size_t pageId, size_t bufSize);
    virtual ~Page();

    // access
    size_t      page_id() const         { return _pageId; }
    size_t      data_len() const        { return _dataLen; }
    size_t      prev_page() const       { return (_pageId == 0) ? size_t(-1) : _prevPage; }
    size_t      next_page() const       { return (_pageId == 0) ? size_t(-1) : _nextPage; }
    bool        is_dirty() const        { return _dirty; }
    size_t      num_keys() const        { return _keyList.size(); }

    // update
    KeyList&    key_list()              { return _keyList; }

    void        data_len(size_t len)    { _dataLen = len; _dirty = true; }
    void        prev_page(size_t pid)   { _prevPage = pid; _dirty = true; }
    void        next_page(size_t pid)   { _nextPage = pid; _dirty = true; }

    // test
    bool        must_split() const;

    // access / update the key list
    std::pair<size_t,bool>
                add_key(BtreeIndex&, const value_type&);    // returns the slot of the new key

    void        remove_key(size_t slot);
    void        update_key(size_t slot, size_t link);

    void        insert_page_link(typename KeyList::iterator kit, Page* p);
    void        update_page_link(Page* p);
    void        remove_page_link(Page* p);  // returns same as remove_key

    // virtual functions
    virtual Page*       split(BtreeIndex&, Iterator&);  // returns the new page
    virtual void        compact(BtreeIndex&, Iterator&);
    virtual oXDR_Stream& encode(oXDR_Stream&);
    virtual iXDR_Stream& decode(iXDR_Stream&);
private:
    Page(const Page&);
    Page& operator=(const Page&);
}; //> BtreeIndex::Page


//
// BtreeIndex::RootPage
//
template<typename Key, typename Compare, typename KeyTraits>
class BtreeIndex<Key,Compare,KeyTraits>::RootPage : public Page {   
    // data
    size_t      _pageSize;
    bool        _uniqueKeys;
    size_t      _totalKeys;
    short       _depth;
    bool        _dirtyTotal;
public:
    RootPage(iXDR_Stream& xstrm);
    RootPage(size_t pageSize, bool uniqueKeys);
    virtual ~RootPage();

    // access
    size_t      page_size() const       { return _pageSize; }
    bool        unique_keys() const     { return _uniqueKeys; }
    size_t      total_keys() const      { return _totalKeys; }
    short       depth() const           { return _depth; }
    size_t      next_free_page() const  { return _nextPage; }
    bool        is_total_dirty() const  { return _dirtyTotal; }

    // update
    void        incr_total()            { ++_totalKeys; _dirtyTotal = true; }
    void        decr_total()            { --_totalKeys; _dirtyTotal = true; }
    void        add_free_page(Page* p)  { p->next_page(next_page()); next_page(p->page_id()); }
    void        rem_free_page(Page* p)  { next_page(p->next_page()); }

    // virtual functions
    virtual Page*       split(BtreeIndex&, Iterator&);
    virtual void        compact(BtreeIndex&, Iterator&);
    virtual oXDR_Stream& encode(oXDR_Stream&);
    virtual iXDR_Stream& decode(iXDR_Stream&);
}; //> BtreeIndex::RootPage


// class which handles read/write to/from an externally specified memory area
template<typename Key, typename Compare, typename KeyTraits>
class BtreeIndex<Key,Compare,KeyTraits>::XIODataStreambuf : public XDR_Streambuf {
public:
    XIODataStreambuf(XDR_Char* buf, size_t size, size_t len = 0);
    ~XIODataStreambuf();

    virtual pos_type seekoff(off_type off, 
                             std::ios_base::seekdir dir, 
                             std::ios_base::openmode which);
    virtual pos_type seekpos(pos_type pos, std::ios_base::openmode which);
};


template<typename Key, typename Compare, typename KeyTraits>
class BtreeIndex<Key,Compare,KeyTraits>::XOutDataBuf : public oXDR_Stream {
    // data
    oXDR_Stream&        _refstrm;
    XIODataStreambuf    _xstreambuf;

public:
    XOutDataBuf(oXDR_Stream&, XDR_Char* buf, size_t size);
    ~XOutDataBuf();
};


template<typename Key, typename Compare, typename KeyTraits>
class BtreeIndex<Key,Compare,KeyTraits>::XInDataBuf : public iXDR_Stream {
    // data
    iXDR_Stream&        _refstrm;
    XIODataStreambuf    _xstreambuf;

public:
    XInDataBuf(iXDR_Stream&, XDR_Char* buf, size_t len);
    ~XInDataBuf();
};


//
// BtreeIndex::Iterator - inline function definitions
//
template<typename Key, typename Compare, typename KeyTraits>
inline typename BtreeIndex<Key,Compare,KeyTraits>::Iterator&
BtreeIndex<Key,Compare,KeyTraits>::Iterator::operator++()
{ _btree->next_iter(*this); return *this; }

template<typename Key, typename Compare, typename KeyTraits>
inline typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::Iterator::operator++(int)
{ Iterator t = *this; _btree->next_iter(*this); return t; }

template<typename Key, typename Compare, typename KeyTraits>
inline typename BtreeIndex<Key,Compare,KeyTraits>::Iterator&
BtreeIndex<Key,Compare,KeyTraits>::Iterator::operator--()
{ _btree->prev_iter(*this); return *this; }

template<typename Key, typename Compare, typename KeyTraits>
inline typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::Iterator::operator--(int)
{ Iterator t = *this; _btree->prev_iter(*this); return t; }

template<typename Key, typename Compare, typename KeyTraits>
inline bool
BtreeIndex<Key,Compare,KeyTraits>::Iterator::operator==(const Iterator& rhs) const
{ return _btree == rhs._btree 
            and _page == rhs._page 
                and _slot == rhs._slot; }

template<typename Key, typename Compare, typename KeyTraits>
inline bool 
BtreeIndex<Key,Compare,KeyTraits>::Iterator::operator!=(const Iterator& rhs) const
{ return !(*this == rhs); }

template<typename Key, typename Compare, typename KeyTraits>
inline typename BtreeIndex<Key,Compare,KeyTraits>::Iterator::value_type
BtreeIndex<Key,Compare,KeyTraits>::Iterator::get_value()
{
    value_type val;
    _btree->get_value(*this, val);
    return val;
}

template<typename Key, typename Compare, typename KeyTraits>
inline void
BtreeIndex<Key,Compare,KeyTraits>::Iterator::update_link(size_t link)
{
    _btree->update_link(*this, link);
}

template<typename Key, typename Compare, typename KeyTraits>
inline BtreeIndex<Key,Compare,KeyTraits>::Iterator::Iterator(
            BtreeIndex* btree, 
            size_t page, 
            size_t slot)
  : _btree(btree), _page(page), _slot(slot) {}


//
// BtreeIndex inline function definitions
//
template<typename Key, typename Compare, typename KeyTraits>
inline bool
BtreeIndex<Key,Compare,KeyTraits>::is_open()
{ return _xstream.is_open(); }


template<typename Key, typename Compare, typename KeyTraits>
inline 
BtreeIndex<Key,Compare,KeyTraits>::operator void*() const
{
    if (_xstream.fail()) return null;
    return const_cast<BtreeIndex*>(this);
}


template<typename Key, typename Compare, typename KeyTraits>
inline bool
BtreeIndex<Key,Compare,KeyTraits>::operator!() const
{ return _xstream.fail(); }


template<typename Key, typename Compare, typename KeyTraits>
inline bool
BtreeIndex<Key,Compare,KeyTraits>::empty()
{ return size() == 0; }


template<typename Key, typename Compare, typename KeyTraits>
inline size_t
BtreeIndex<Key,Compare,KeyTraits>::size()
{ return _root->total_keys(); }


template<typename Key, typename Compare, typename KeyTraits>
inline std::pair<typename BtreeIndex<Key,Compare,KeyTraits>::Iterator, bool>
BtreeIndex<Key,Compare,KeyTraits>::insert(const Key& key)
{ return insert(value_type(key, 0)); }


template<typename Key, typename Compare, typename KeyTraits>
inline Compare
BtreeIndex<Key,Compare,KeyTraits>::key_comp() const
{ return _comp; }


template<typename Key, typename Compare, typename KeyTraits>
inline std::string
BtreeIndex<Key,Compare,KeyTraits>::file_name() const
{ return _name; }


template<typename Key, typename Compare, typename KeyTraits>
inline size_t
BtreeIndex<Key,Compare,KeyTraits>::page_size() const
{ return _pageSize; }


template<typename Key, typename Compare, typename KeyTraits>
inline bool
BtreeIndex<Key,Compare,KeyTraits>::unique_keys() const
{ return _uniqueKeys; }


template<typename Key, typename Compare, typename KeyTraits>
inline bool
BtreeIndex<Key,Compare,KeyTraits>::read_only() const
{ return _readOnly; }


template<typename Key, typename Compare, typename KeyTraits>
inline XDR_Stream&
BtreeIndex<Key,Compare,KeyTraits>::xdrstream()
{ return _xstream; }


template<typename Key, typename Compare, typename KeyTraits>
inline typename BtreeIndex<Key,Compare,KeyTraits>::DeleteStrategy
BtreeIndex<Key,Compare,KeyTraits>::set_delete_strategy(DeleteStrategy s)
{
    DeleteStrategy t = _dStrategy;
    _dStrategy = s;
    return t;
}

template<typename Key, typename Compare, typename KeyTraits>
inline typename BtreeIndex<Key,Compare,KeyTraits>::UpdateStrategy
BtreeIndex<Key,Compare,KeyTraits>::set_update_strategy(UpdateStrategy s)
{
    UpdateStrategy t = _uStrategy;
    _uStrategy = s;
    return t;
}


template<typename Key, typename Compare, typename KeyTraits>
inline int
BtreeIndex<Key,Compare,KeyTraits>::set_cache_size(int size)
{
    int t = _pageCache.max_size();
    _pageCache.max_size(size);
    return t;
}


} //> namespace


#ifdef EXPORT_NOT_SUPPORTED
#include "BtreeIndex.hxx"
#include "BtreeIndex_Page.hxx"
#include "BtreeIndex_RootPage.hxx"
#include "BtreeIndex_PageCache.hxx"
#include "BtreeIndex_XIODataBuf.hxx"
#endif


#endif
