/*******************************************************************************
  File: BtreeIndex.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
    The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "BtreeIndex.h"

// Standard C++ headers
#include <sstream>

// Standard C headers
#include <time.h>
#include <stdio.h>


namespace Util {


//
// BtreeIndex functions
//
template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex<Key,Compare,KeyTraits>::BtreeIndex(Compare comp)
  : _comp(comp),
    _name(),
    _pageSize(2048),
    _uniqueKeys(true),
    _readOnly(false),
    _xstream(),
    _root(null),
    _pageCache(_xstream),
    _dStrategy(Compact),
    _uStrategy(MostlySafe)
{
    // empty
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex<Key,Compare,KeyTraits>::BtreeIndex(
            const std::string& name, 
            bool readOnly,
            Compare comp)
  : _comp(comp),
    _name(name),
    _pageSize(2048),
    _uniqueKeys(true),
    _readOnly(readOnly),
    _xstream(),
    _root(null),
    _pageCache(_xstream),
    _dStrategy(Compact),
    _uStrategy(MostlySafe)
{
    open(name, readOnly);
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex<Key,Compare,KeyTraits>::~BtreeIndex()
{
    close();
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::create(const std::string& name, bool uniqueKeys)
{
    create(name, _pageSize, uniqueKeys);
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::create(const std::string& name,
                                          int pageSize,
                                          bool uniqueKeys)
{
    if (is_open())
        throw std::ios_base::failure("BtreeIndex::create: file already open");
    _name       = name;
    _pageSize   = pageSize;
    _uniqueKeys = uniqueKeys;

    // create a new, empty file
    _xstream.open(name, std::ios_base::out);    // create new file

    // put a root page in it
    RootPage* p = new RootPage(_pageSize, _uniqueKeys);
    p->encode(_xstream);
    _xstream.close();
    delete p;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::open(bool readOnly)
{
    open(_name, readOnly);
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::open(const std::string& name, bool readOnly)
{
    if (is_open())
        throw std::ios_base::failure(
            "BtreeIndex::open: file is already open");

    _name = name;
    _readOnly = readOnly;
    if (readOnly)
        _xstream.open(name, std::ios_base::in);
    else
        _xstream.open(name, std::ios_base::in|std::ios_base::out);
    if (not _xstream.good()) 
        throw std::ios_base::failure(
            "BtreeIndex::open: could not open file");

    // load the root page
    _root = new RootPage(_xstream);
    if (_xstream.eof()) {
        delete _root;
        throw std::ios_base::failure(
            "BtreeIndex::open: file does not appear to be a valid index");
    }

    // Get the page_size & unique keys flags from the root page
    _pageSize   = _root->page_size();
    _uniqueKeys = _root->unique_keys();

    // Set up the page cache
    _pageCache.min_size(_root->depth());

}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::close()
{
    if (not is_open()) return;
    flush();            // write out all dirty pages
    _pageCache.clear(); // delete all cached pages
    delete _root;       // ditto for the root
    _xstream.close();
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::flush()
{
    flush(Safe);
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::begin()
{
    if (empty()) return end();
    return lower_bound( _root->key_list().front().key );
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::end()
{
    return Iterator(this, -1, 0);
}


template<typename Key, typename Compare, typename KeyTraits>
std::pair<typename BtreeIndex<Key,Compare,KeyTraits>::iterator, bool>
BtreeIndex<Key,Compare,KeyTraits>::insert(const value_type& x)
{
    std::pair<Iterator,bool> rtn;
    rtn = insert(_root, _root->depth(), x);
    if (rtn.second) {
        _root->incr_total();
        if (_root->must_split()) {
            _root->split(*this, rtn.first);
            _pageCache.min_size(_root->depth());
        }
    }
    flush(_uStrategy);
    return rtn;
}


#ifndef NESTED_TEMPLATE_DEFINITION_NOT_SUPPORTED
template<typename Key, typename Compare, typename KeyTraits>
template<typename InputIterator>
void
BtreeIndex<Key,Compare,KeyTraits>::insert(InputIterator first, InputIterator last)
{
    while (first != last) {
        insert(_root, _root->depth, *first);
        if (_root->must_split()) _root->split();
        _root->incr_total();
        ++first;
    }
    flush(_uStrategy);
}
#endif


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::iterator
BtreeIndex<Key,Compare,KeyTraits>::erase(iterator pos)
{
    // The 'pos' does not contain the key, but we need it to chase the links
    value_type val = pos.get_value();

    // Find a path to the leaf page and remove the key
    bool ok = erase(_root, _root->depth(), val.first, pos);
    if (ok) {
        _root->decr_total();
        if (_root->must_split()) {
            _root->split(*this, pos);
        } else {
            _root->compact(*this, pos);
        }
        flush(_uStrategy);
    }

    return pos;
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::size_type
BtreeIndex<Key,Compare,KeyTraits>::erase(const Key& x)
{
    size_t count = 0;
    iterator it = lower_bound(x);
    while (it != end() and not _comp(x, it.get_value().first)) {
        it = erase(it);
        ++count;
    }
    return count;
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::iterator
BtreeIndex<Key,Compare,KeyTraits>::erase(iterator first, Iterator last)
{
    size_t count = std::distance(first,last);
    while (count > 0) {
        first = erase(first);
        --count;
    }
    return first;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::clear()
{
    if (_readOnly)
        throw std::ios_base::failure("BtreeIndex::clear: file is opened for read access only");

    // Close the file
    close();
    // Create a new one
    create(_name, page_size(), unique_keys());
    // Reopen it
    open();
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::find(const Key& k)
{
    // Use the internal lower_bound
    if (empty()) return end();
    Iterator it = lower_bound(_root, _root->depth(), k);
    if (it != end() and _comp(k, it.get_value().first)) it = end();
    _pageCache.reset();
    return it;
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::size_type
BtreeIndex<Key,Compare,KeyTraits>::count(const Key& k)
{
    // Find the lower bound
    if (empty()) return 0;
    Iterator it = lower_bound(_root, _root->depth(), k);
    // Now count the keys until the upper bound is found
    size_type cnt = 0;
    while (it != end() and not _comp(k, it.get_value().first)) {
        ++cnt;
        ++it;
    }
    return cnt;
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::lower_bound(const Key& k)
{
    if (empty()) return end();
    Iterator it = lower_bound(_root, _root->depth(), k);
    _pageCache.reset();
    return it;
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::upper_bound(const Key& k)
{
    if (empty()) return end();
    Iterator it = upper_bound(_root, _root->depth(), k);
    _pageCache.reset();
    return it;
}


template<typename Key, typename Compare, typename KeyTraits>
std::pair<typename BtreeIndex<Key,Compare,KeyTraits>::Iterator, 
          typename BtreeIndex<Key,Compare,KeyTraits>::Iterator>
BtreeIndex<Key,Compare,KeyTraits>::equal_range(const Key& x)
{
    return std::make_pair(lower_bound(x), upper_bound(x));
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::reindex()
{
    // Create a temporary to hold the new index
    time_t t = ::time(0);
    std::ostringstream ostrm;
    ostrm << "tmp" << t << ".idx";
    BtreeIndex tidx;
    tidx.create(ostrm.str());
    tidx.open();

    // Copy the index to the temporary
    Iterator it = begin();
    for (; it != end(); ++it) {
        tidx.insert(it.get_value());
    }

    // Close both indecies
    close();
    tidx.close();

    // Remove the current one
    ::remove(file_name().c_str());

    // Rename the new one
    ::rename(tidx.file_name().c_str(), file_name().c_str());

    // Reopen the new one
    open();
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::prev_iter(Iterator& it)
{
    if (it._slot > 0) {
        --it._slot; 
    } else if (it == end()) {
        // Walk down the tree and find the last slot on the last page
        int depth = _root->_depth;
        Page* p = _root;
        while (depth > 0) {
            p = get_page(p->key_list().back().link);
            --depth;
        }
        it._page = p->page_id();
        it._slot = p->num_keys() - 1;
    } else {
        Page* p = get_page(it._page);
        Page* prevp = get_page(p->prev_page());
        if (prevp) {
            it._page = prevp->page_id();
            it._slot = prevp->num_keys() - 1;
        }
    }
    _pageCache.reset();
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::next_iter(Iterator& it)
{
    Page* p = get_page(it._page);
    if (p) {
        if (++it._slot < p->num_keys()) return;
        else {
            it._page = p->next_page();
            it._slot = 0;
        }
    }
    _pageCache.reset();
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::get_value(Iterator& it, value_type& value)
{
    Page* p = get_page(it._page);
    if (not p or it._slot >= p->num_keys())
        throw std::ios_base::failure(
            std::string("BtreeIndex::get_value: invalid iterator dereference"));
        
    // Find the slot
    typename Page::KeyList::iterator kit = p->key_list().begin();
    for (int ii = 0; size_t(ii) < it._slot; ++ii)
        ++kit;
    value = std::make_pair((*kit).key, (*kit).link);
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::update_link(Iterator& it, size_t link)
{
    Page* p = get_page(it._page);
    p->update_key(it._slot, link);
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Page*
BtreeIndex<Key,Compare,KeyTraits>::get_page(size_t pageId)
{
    if (pageId == size_t(-1)) return null;
    else if (pageId == 0) return _root;
    else return _pageCache.get_page(pageId, _pageSize);
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Page*
BtreeIndex<Key,Compare,KeyTraits>::get_free_page()
{
    Page* p = get_page(_root->next_free_page());
    if (p) {
        _root->rem_free_page(p);
    } else {
        _xstream.seekp(0, std::ios_base::end);
        size_t pageId = _xstream.tellp();
        p = new Page(pageId, _pageSize);
        p->encode(_xstream); // flush it so the next call works correctly
        _pageCache.add_page(p);
    }
    return p;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::add_free_page(Page* p)
{
    _root->add_free_page(p);
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::lower_bound(Page* p, int depth, const Key& key)
{
    if (depth == 0) {
        typename Page::KeyList& kl = p->key_list();
        typename Page::KeyList::iterator it = kl.begin();
        size_t slot = 0;
        for (; it != kl.end(); ++it, ++slot) {
            if ( not _comp((*it).key, key) ) break;
        }
        if (it != kl.end()) {
            return Iterator(this, p->page_id(), slot);
        } else if (p->next_page() == size_t(-1)) {
            return end();
        } else {
            return Iterator(this, p->next_page(), 0);
        }
    } else {
        // find the next page to look on
        typename Page::KeyList& kl = p->key_list();
        typename Page::KeyList::iterator it = kl.begin();
        for (++it; it != kl.end(); ++it) {
            if ( not _comp((*it).key, key) ) break;
        }
        --it;
        return lower_bound(get_page((*it).link), depth-1, key);
    }
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Iterator
BtreeIndex<Key,Compare,KeyTraits>::upper_bound(Page* p, int depth, const Key& key)
{
    if (depth == 0) {
        typename Page::KeyList& kl = p->key_list();
        typename Page::KeyList::iterator it = kl.begin();
        size_t slot = 0;
        for (; it != kl.end(); ++it, ++slot) {
            if ( _comp(key, (*it).key) ) break;
        }
        if (it != kl.end()) {
            return Iterator(this, p->page_id(), slot);
        } else if (p->next_page() == size_t(-1)) {
            return end();
        } else {
            return Iterator(this, p->next_page(), 0);
        }
    } else {
        // find the next page to look on
        typename Page::KeyList& kl = p->key_list();
        typename Page::KeyList::iterator it = kl.begin();
        for (++it; it != kl.end(); ++it) {
            if ( _comp(key, (*it).key) ) break;
        }
        --it;
        return upper_bound(get_page((*it).link), depth-1, key);
    }
}


template<typename Key, typename Compare, typename KeyTraits>
std::pair<typename BtreeIndex<Key,Compare,KeyTraits>::iterator, bool>
BtreeIndex<Key,Compare,KeyTraits>::insert(Page* p, int depth, const value_type& x)
{
    if (depth == 0) {   // leaf page
        std::pair<size_t,bool> rtn;
        rtn = p->add_key(*this, x);
        return std::make_pair(Iterator(this, p->page_id(), rtn.first), rtn.second);
    } else {            // internal page
        std::pair<Iterator,bool> rtn;
        Page* nextp;
        typename Page::KeyList& kl = p->key_list();
        typename Page::KeyList::iterator kit = kl.begin();
        if ( _comp(x.first, (*kit).key) ) {
            // Boundary condition - key is less than first key on this page
            nextp = get_page((*kit).link);
            rtn = insert(nextp, depth-1, x);
            p->update_page_link(nextp);
        } else {
            // find which page to put the new key in
            for (++kit; kit != kl.end(); ++kit) {
                if ( _comp(x.first, (*kit).key) ) break;
            }
            --kit;
            nextp = get_page((*kit).link);
            rtn = insert(nextp, depth-1, x);
        }
        if (nextp->must_split()) {
            Page* newp = nextp->split(*this, rtn.first);
            p->insert_page_link(++kit, newp);
        }
        return rtn;
    }
}


template<typename Key, typename Compare, typename KeyTraits>
bool
BtreeIndex<Key,Compare,KeyTraits>::erase(Page* p, int depth, const Key& key, iterator& pos)
{
    using std::pair;
    using std::make_pair;

    // we have to find the correct path to the page in pos
    if (depth == 0) {
        if (pos._page != p->page_id()) return false;
        // Ok - remove the key
        p->remove_key(pos._slot);
        if (pos._slot >= p->num_keys()) {
            pos._page = p->next_page();
            pos._slot = 0;
        }
        return true;
    } else {
        // find the link we think will lead to pos
        typename Page::KeyList& kl = p->key_list();
        typename Page::KeyList::iterator kit = kl.begin();
        for (++kit; kit != kl.end(); ++kit) {
            if (not _comp((*kit).key, key)) break;  // use lower bound test
        }
        --kit;  // back up one link
        Page* nextp = get_page((*kit).link);
        bool removed = false;
        for (;;) {
            removed = erase(nextp, depth-1, key, pos);
            if (removed) break;
            ++kit;
            if (kit == kl.end()) return removed;
            nextp = get_page((*kit).link);
        }
        
        if (not removed)
            throw std::ios_base::failure(
                std::string("BtreeIndex::erase(2): invalid iterator (page not found)"));

        // we got to the right page and updated it
        // now see what we have to do at this level
        if (nextp->num_keys() == 0) {
            add_free_page(nextp);
            p->remove_page_link(nextp);
            return true;
        }

        // See if the page link is still correct
        if ( _comp((*kit).key, nextp->key_list().front().key) ) {
            // we removed or updated the first key in the lower page
            p->update_page_link(nextp); // might grow the page
        }

        if (nextp->must_split()) {
            // the lower page might have grown
            Page* newp = nextp->split(*this, pos);
            p->insert_page_link(++kit, newp);
        } else if (_dStrategy == Compact) {
            p->compact(*this, pos);
        }
        return true;
    }
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::flush(UpdateStrategy s)
{
    _pageCache.reset(); // remove excess pages
    if (s == Safe) {
        _pageCache.flush();
        if (_root->is_total_dirty() or _root->is_dirty())
            _root->encode(_xstream);
    } else if (s == MostlySafe) {
        _pageCache.flush();
        if (_root->is_dirty()) _root->encode(_xstream);
    }
}


} //> namespace

