/*******************************************************************************
  File: BtreeIndex_PageCache.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
    The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "IndexedFile.h"

namespace Util {


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex<Key,Compare,KeyTraits>::PageCache::PageCache(XDR_Stream& xstrm)
  : _xstream(xstrm),
    _minSize(0),
    _maxSize(0),
    _pageList()
{
    // empty
}


template<typename Key, typename Compare, typename KeyTraits>
BtreeIndex<Key,Compare,KeyTraits>::PageCache::~PageCache()
{
    flush();
    clear();
}


template<typename Key, typename Compare, typename KeyTraits>
typename BtreeIndex<Key,Compare,KeyTraits>::Page*
BtreeIndex<Key,Compare,KeyTraits>::PageCache::get_page(size_t pid, size_t pgSize)
{
    // See if the page is in the cache
    typename PageList::iterator pit = _pageList.begin();
    for (; pit != _pageList.end(); ++pit) {
        if ((*pit)->page_id() == pid) break;
    }
    Page* p = 0;
    if (pit != _pageList.end()) {   
        // Ok - its in the cache, move it to the front
        p = (*pit);
        _pageList.splice(_pageList.begin(), _pageList, pit);
    } else { 
        // Fetch the page from external store
        p = new Page(pid, pgSize);
        p->decode(_xstream);
        _pageList.push_front(p);
    }
    return p;
}


template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::PageCache::add_page(Page* page)
{
    _pageList.push_front(page);
}

template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::PageCache::reset()
{
    // Remove any excess pages
    int siz = std::max(_minSize, _maxSize);
    while (_pageList.size() > size_t(siz)) {
        Page* p = _pageList.back();
        if (p->is_dirty()) p->encode(_xstream);
        delete p;
        _pageList.pop_back();
    }
}

template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::PageCache::clear()
{
    // Remove all pages
    while (_pageList.size() > 0) {
        Page* p = _pageList.front();
        delete p;
        _pageList.pop_front();
    }
}

template<typename Key, typename Compare, typename KeyTraits>
void
BtreeIndex<Key,Compare,KeyTraits>::PageCache::flush()
{
    typename PageList::iterator pit = _pageList.begin();
    for (; pit != _pageList.end(); ++pit) {
        if ((*pit)->is_dirty()) (*pit)->encode(_xstream);
    }
}


} //> namespace

