/*******************************************************************************
  File: IndexedFile.hxx
  Copyright (c) 2002 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, such as copyrighting
  or claiming authorship, to prevent the continued distribution of this 
  Software under an open source distribution model.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
    The template implementation file for the BtreeIndex class.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/

// My header
#include "IndexedFile.h"


namespace Util {


//
// IndexedFile::Iterator -
//
template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator::get_value(value_type& value)
{
    typename index_type::value_type ival = _base.get_value();
    value.first = ival.first;
    _ifile->read(ival.second, value.second); 
}


template<typename Key, typename Item, typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator::update_item(Item& item)
{
    size_t pos = _ifile->write(item);
    _base.update_link(pos);
}


//
// IndexedFile -
//
template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::IndexedFile(Compare comp)
  : _name(),
    _index(comp),
    _xstream()
{
    // empty
}


template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::IndexedFile(
            const std::string& name, 
            bool readOnly,
            Compare comp)
  : _name(name),
    _index(comp),
    _xstream()
{
    open(name, readOnly);
}


template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::~IndexedFile()
{
    close();
}


template<typename Key, typename Item,
         typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::create(
        const std::string& name, bool uniqueKeys)
{
    create(name, _index.page_size(), uniqueKeys);
}


template<typename Key, typename Item,
         typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::create(
        const std::string& name, int pageSize, bool uniqueKeys)
{
    const char* const func = "IndexedFile::create";
    if (is_open())
        throw std::ios_base::failure(
            std::string(func).append(": file already open"));
    _name = name;

    // Add '.xi' to the name for the index file
    _index.create(name + ".xi", pageSize, uniqueKeys);

    // Add '.xd' to the name for the data file
    std::string fname = name + ".xd";
    _xstream.open(fname, std::ios_base::out);   // create a new, empty file
    _xstream.close();
}


template<typename Key, typename Item,
         typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::open(bool readOnly)
{
    open(_name, readOnly);
}


template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::open(
            const std::string& name,
            bool readOnly)
{
    const char* const func = "IndexedFile::open";
    if (is_open())
        throw std::ios_base::failure(
            std::string(func).append(": file already open"));
    _name = name;

    // Add '.xi' to the name for the index file
    _index.open(name + ".xi", readOnly);

    // Add '.xd' to the name for the data file
    std::string fname = name + ".xd";
    if (readOnly)
        _xstream.open(fname, std::ios_base::in);
    else
        _xstream.open(fname, std::ios_base::in|std::ios_base::out);
    if (not _xstream.good())
        throw std::ios_base::failure(
            std::string(func).append(": can not open data file"));
}


template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::close()
{
    _index.close();
    _xstream.close();
}


template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::flush()
{
    _index.flush();
    _xstream.flush();
}


template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
std::pair<typename IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::Iterator, bool>
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::insert(const value_type& x)
{
    // Write the data to the data file
    size_t link = write(x.second);
    
    // Add the key to the index
    std::pair<typename index_type::iterator,bool> rtn = 
        _index.insert( typename index_type::value_type(x.first, link) );

    return std::make_pair(Iterator(this, rtn.first), rtn.second);
}



#ifndef NESTED_TEMPLATE_DEFINITION_NOT_SUPPORTED
template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
template<typename InputIterator>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::insert(
            InputIterator first,
            InputIterator last)
{
    while (first != last) {
        insert(*first);
        ++first;
    }
}
#endif


template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
void
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::read(size_t pos, Item& item)
{
    _xstream.seekg(pos);
    ItemTraits::decode(_xstream, item);
}


template<typename Key, typename Item, 
         typename Compare, typename KeyTraits, typename ItemTraits>
size_t
IndexedFile<Key,Item,Compare,KeyTraits,ItemTraits>::write(const Item& item)
{
    _xstream.seekp(0, std::ios_base::end);
    size_t link = _xstream.tellp();
    ItemTraits::encode(_xstream, item);
    return link;
}


} //> namespace


