/*******************************************************************************
  File: XIO.h
  Copyright (c) 1999 Bleading Edge Software Technologies, Inc.

  This Software is distributed under the open source, free software model.
  Permission to use, modify, and distribute this Software, including
  incorporating it into proprietary software, is granted without fee, 
  provided this copyright notice appear in all copies of the code.
  You are under no obligation to distribute source code which is derived from
  or uses this Software. You may not do anything however, to prevent the 
  continued distribution of this Software under an open source distribution
  model. This includes, but is not limited to, such things as copyrighting
  or claiming authorship of the Software.
  This Software is distributed in the hope that it will be useful, but it is
  provide "as is", WITHOUT ANY WARRANTY; without even the implied warranty
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

//$-----------------------------------------------------------------------------
  Author: Jack W. Reeves

//!----------------------------------------------------------------------------|
Description:
  Define both file based and string based XDR_Streams. Base them on the 
  regular fstream and stringstream classes.
  
//#----------------------------------------------------------------------------|
Notes:
  
//%****************************************************************************/
#ifndef UTIL_XIO_H
#define UTIL_XIO_H

#include "platform.h"

#include "XDRStream.h"

// Standard C++ headers
#include <fstream>
#include <sstream>

namespace Util {

//
// File base XDR_Streams
//
class XIOFilebuf : public XDR_Streambuf {
    // typedefs
    typedef XDR_Streambuf                   inherited;
    typedef inherited::traits_type          traits;
    typedef inherited::int_type             int_type;
    typedef traits::pos_type                pos_type;
    typedef traits::off_type                off_type;
    typedef std::streamsize                 streamsize;

    // data
    std::filebuf        _fb;
    char_type           _ibuf[128];     // so we can make underflow / uflow work properly
    char_type           _obuf[128];     // just to make things more efficient

public:
    XIOFilebuf();
    virtual ~XIOFilebuf();

    std::filebuf*           rdbuf() const   { return const_cast<std::filebuf*>(&_fb); }

    bool                    is_open();
    XIOFilebuf*             open(const std::string& name, std::ios_base::openmode mode);
    XIOFilebuf*             close();
    
protected:
    virtual inherited*      setbuf(char* s, streamsize n);
    virtual pos_type        seekoff(off_type off, std::ios_base::seekdir dir,
                                    std::ios_base::openmode which = std::ios_base::in | std::ios_base::out);
    virtual pos_type        seekpos(pos_type pos,
                                    std::ios_base::openmode which = std::ios_base::in | std::ios_base::out);
    virtual int             sync();
    
    virtual int_type        underflow();
    
    virtual int_type        overflow(int_type c = traits::eof());
};


class iXIOFileStream : public iXDR_Stream {
    // data
    XIOFilebuf          _xfb;
    
public:
    iXIOFileStream();
    iXIOFileStream(const std::string& name, std::ios_base::openmode mode = std::ios_base::in);
    ~iXIOFileStream();

    XIOFilebuf*     rdbuf() const   { return const_cast<XIOFilebuf*>(&_xfb); }

    bool            is_open();
    void            open(const std::string& name, std::ios_base::openmode mode = std::ios_base::in);
    void            close();
    
private:
    iXIOFileStream(const iXIOFileStream&);
    iXIOFileStream& operator=(const iXIOFileStream&);
};
    


class oXIOFileStream : public oXDR_Stream {
    // data
    XIOFilebuf              _xfb;
    
public:
    oXIOFileStream();
    oXIOFileStream(const std::string& name, std::ios_base::openmode mode = std::ios_base::out);
    virtual ~oXIOFileStream();

    XIOFilebuf*     rdbuf() const   { return const_cast<XIOFilebuf*>(&_xfb); }

    bool            is_open();
    void            open(const std::string& name, std::ios_base::openmode mode = std::ios_base::out);
    void            close();
    
private:
    oXIOFileStream(const oXIOFileStream&);
    oXIOFileStream& operator=(const oXIOFileStream&);
};
    

class XIOFileStream : public XDR_Stream {
    // data
    XIOFilebuf          _xfb;
    
public:
    XIOFileStream();
    XIOFileStream(const std::string& name, std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out);
    virtual ~XIOFileStream();

    XIOFilebuf*     rdbuf() const   { return const_cast<XIOFilebuf*>(&_xfb); }

    bool            is_open();
    void            open(const std::string& name, std::ios_base::openmode mode = std::ios_base::in);
    void            close();
    
private:
    XIOFileStream(const XIOFileStream&);
    XIOFileStream& operator=(const XIOFileStream&);
};


//
// A string based XDRStream
//
class XIOStringbuf : public XDR_Streambuf {
    // typedefs
    typedef XDR_Streambuf                   inherited;
    typedef inherited::traits_type          traits;
    typedef inherited::int_type             int_type;
    typedef traits::pos_type                pos_type;
    typedef traits::off_type                off_type;
    typedef std::streamsize                 streamsize;

    // data
    std::stringbuf      _sb;
    char_type           _ibuf[128];     // so we can make underflow / uflow work properly
    char_type           _obuf[128];     // just to make things more efficient

public:
    XIOStringbuf(std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out);
    explicit XIOStringbuf(const std::string& str, std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out);
    virtual ~XIOStringbuf();

    std::stringbuf*         rdbuf() const   { return const_cast<std::stringbuf*>(&_sb); }

    std::string             str() const;
    void                    str(const std::string& s);
    
protected:
    virtual inherited*      setbuf(char* s, streamsize n);
    virtual pos_type        seekoff(off_type off, std::ios_base::seekdir dir,
                                    std::ios_base::openmode which = std::ios_base::in | std::ios_base::out);
    virtual pos_type        seekpos(pos_type pos,
                                    std::ios_base::openmode which = std::ios_base::in | std::ios_base::out);
    virtual int             sync();
    
    virtual int_type        underflow();
    
    virtual int_type        overflow(int_type c = traits::eof());
};



class XIOStringStream : public XDR_Stream {
    // data
    XIOStringbuf            _xsb;
    
public:

    XIOStringStream(std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out);
    XIOStringStream(const std::string& str, std::ios_base::openmode mode = std::ios_base::in | std::ios_base::out);
    virtual ~XIOStringStream();

    XIOStringbuf*   rdbuf() const   { return const_cast<XIOStringbuf*>(&_xsb); }

    std::string             str() const;
    void                    str(const std::string& s);

private:
    XIOStringStream(const XIOStringStream&);
    XIOStringStream& operator=(const XIOStringStream&);
};



} //> namespace Util

#endif
