/*
   +------------------------+
   | Cleanup Framework in C |`
   | ~~~~~~~~~~~~~~~~~~~~~~ |`
   | Stephane GRAD          |`
   | (2001)                 |`
   +------------------------+`
    ``````````````````````````

   +---------+-------+------------+----------------+------------+
   | Defines | Types | Prototypes | Initialization | Guidelines |
---+---------+-------+------------+----------------+------------+---------------
*/

/*
   +---------+
   | Defines |
---+---------+------------------------------------------------------------------
*/

/* should be passed as arg to the compiler
 */
//#define DEBUGCL



/* helpers for debug information
 */
#ifdef DEBUGCL

#define DEBUGCL_EXTRA_PARAMS \
, const char* aFileName, const int aLineNo

#define DEBUGCL_EXTRA_ARGS \
, aFileName, aLineNo

#define DEBUGCL_EXTRA_INFO \
, __FILE__, __LINE__

#define DEBUGCL_EXTRA_STRUCT \
const char* filename; \
int lineno;

#else

#define DEBUGCL_EXTRA_PARAMS
#define DEBUGCL_EXTRA_ARGS
#define DEBUGCL_EXTRA_INFO
#define DEBUGCL_EXTRA_STRUCT

#endif /* DEBUGCL */

/* push on the current (local) stack
 */
#define PUSHCL(ptr, clean) _cl_push(&_cl_cleanup_stack, (ptr), (clean) DEBUGCL_EXTRA_INFO)

/* pop on the current (local) stack
 */
#define POPCL(n) _cl_pop(&_cl_cleanup_stack, (n) DEBUGCL_EXTRA_INFO)

/* define a global stack to avoid errors on 'return' keyword
 * if there is no cleanup stack in the local context
 * and 'return' is used
 */
#define INITCL_GLOBAL() \
static tCleanupStack _cl_cleanup_stack = { 0, 0, 0 DEBUGCL_EXTRA_INFO }

/* define a stack for the current context
 */
#define INITCL(n) \
tCleanupElm _cl_current_stack[(n)]; \
tCleanupStack _cl_cleanup_stack = { (n), _cl_current_stack, 0 DEBUGCL_EXTRA_INFO }; \
if (0) goto _cl_l_cleanup
/* goto _cl_l_cleanup is there to raise a compile-time error
 * if the user forgets to put DONECL() at the end
 * of its function
 */

/* should be the end of the stack life
 * in the current context
 */
#define DONECL() \
_cl_l_cleanup: \
_cl_popa(&_cl_cleanup_stack DEBUGCL_EXTRA_INFO)
/* if the user does not use 'return' keyword
 * at the end of its function, the stack will be pop'ed
 * anyway
 */

/* overload 'return' keyword in order to pop the stack
 * prior to leaving the function
 */
#define return {_cl_popa(&_cl_cleanup_stack DEBUGCL_EXTRA_INFO);} return

/*
             +-------+
             | Types |
-------------+-------+----------------------------------------------------------

*/

/* element of a cleanup stack
 */
struct sCleanupElm
{
    void (*clean)(void*); /* cleanup callback */
    void* ptr;            /* pointer to clean */
    DEBUGCL_EXTRA_STRUCT
};
typedef struct sCleanupElm tCleanupElm;

/* cleanup stack structure
 */
struct sCleanupStack
{
    const unsigned int max; /* max number of elements in the stack */
    tCleanupElm* stack;     /* pointer to the stack */
    unsigned int index;     /* current index in the stack */
    DEBUGCL_EXTRA_STRUCT
};
typedef struct sCleanupStack tCleanupStack;

/*
                     +------------+
                     | Prototypes |
---------------------+------------+---------------------------------------------
*/

/* push a pointer with its cleanup callback
 * on a specified stack
 */
void _cl_push(tCleanupStack* aStack, void* aPtr, void (*aClean)(void*)
DEBUGCL_EXTRA_PARAMS);

/* pop n elements out of the specified stack
 */
void _cl_pop(tCleanupStack* aStack, unsigned int n
DEBUGCL_EXTRA_PARAMS);

/* pop all the elements out of a specified stack
 */
void _cl_popa(tCleanupStack* aStack
DEBUGCL_EXTRA_PARAMS);

/*
                                  +----------------+
                                  | Initialization |
----------------------------------+----------------+----------------------------
*/

/* Let's use it here (so this is transparent for the user) */
INITCL_GLOBAL();

/*
                                                   +------------+
                                                   | Guidelines |
---------------------------------------------------+------------+---------------
*/

/*
 * 1. cleanup.h should be included only in .c files (so not in headers)
 *    at the end of the #include list, for performance reasons
 *
 * 2. INITCL(x) / DONECL() cannot be nested in the same function
 *
 * 3. if you want extra information related to stack errors,
 *    #define DEBUGCL in cleanup.h and cleanup.c (preferably via arg to compiler),
 *    and #define DEBUGCL_PRINT in cleanup.c file (before #include cleanup.h)
 */
