//
//      quantity.hpp - rev 0.4 - 2001-07-16
//
//	Copyright (c) 2001 by Michael S. Kenniston.  For the most
//	recent version check www.xnet.com/~msk/quantity.  Permission is granted
//	to use this code without restriction so long as this copyright
//	notice appears in all source files.

//	This code is provided as-is, with no warrantee of correctness.

//	Unless otherwise specified, the definitions of all units in this
//	file are from NIST Special Publication 811, found online at
//	http://physics.nist.gov/Document/sp811.pdf, or from other NIST
//	web pages.

//	TO DO:
//	- templatize ostream (beware gnu compatability)
//  - check on proper way to compute double long pow( x, y )
//  - test with Rep defined to be a high-precision class instead of a primitive
//
// Possible future enhancements:
//	- add fractional dimensions
//	- add any_quantity
//	- templatize on Rep type (float, double, and long double only)
//

#ifndef CUJ_QUANTITY_QUANTITY_HPP_MSK20010704
#define CUJ_QUANTITY_QUANTITY_HPP_MSK20010704

#include <cmath>	// for pow()
#include "boost/config.hpp"

#if ( defined BOOST_MSVC ) || ( defined __MWERKS__ )
	#define CUJ_QUANTITY_NO_PASCAL_IDENTIFIER
#endif

#ifdef CUJ_QUANTITY_NO_PASCAL_IDENTIFIER
	// Some compilers use "pascal" as a keyword.
    // Note that defining a macro "pascal()" with zero arguments
    // will not interfere with the use of "pascal" as a keyword,
    // since the macro will be expanded only when followed by "()".
	#define pascal() pascal_()
#endif

#ifdef BOOST_MSVC // BOOST_MSVC
	// for MSVC, 4248 = invalid access to "private" - must be treated as an error
	// for MSVC, 4786 = truncated names in debugger - ignore
	#pragma warning( error:4248; disable:4786 )
#endif

namespace cuj {
namespace quantity {

#ifdef CUJ_QUANTITY_REP_TYPE
	typedef CUJ_QUANTITY_REP_TYPE Rep;
#else
	typedef double Rep;
#endif

template< typename Dims >
	class quantity; // declare now, define later

// We could drag dimensions around individually, but it's much more convenient to package them.

template< int D1, int D2, int D3, int D4 = 0, int D5 = 0, int D6 = 0, int D7 = 0 >
	struct dimensions
	{
		enum
		{
			dim1 = D1,
			dim2 = D2,
			dim3 = D3,
			dim4 = D4,
			dim5 = D5,
			dim6 = D6,
			dim7 = D7,
            not_all_zero =
                D1 != 0 ||
                D2 != 0 ||
                D3 != 0 ||
                D4 != 0 ||
                D5 != 0 ||
                D6 != 0 ||
                D7 != 0
		};
	};

namespace detail {

	// The "permit" class is a portability hack - anyone with a permit
	// is treated as a friend w.r.t. access to the constructor and
	// m_value accessors of any quantity<>.

	struct permit
	{
		permit() {}
		explicit permit( const Rep & v ) : m_value( v ) {}
        operator Rep() const { return m_value; } // needed so we can construct primitives from it
		Rep m_value;
	};

	// The "collapse" template is used to avoid quantity< dimensions< 0, 0, 0 > >,
	// i.e. to make dimensionless results come out as type "Rep".

	template< typename Dims >
	struct collapse
	{
		typedef quantity< Dims > type;
	};

	template<>
	struct collapse< dimensions< 0, 0, 0 > >
	{
		typedef Rep type;
	};

	// The "static_assert" template is used to trigger a compile-time
    // error if the user attempts to take a root of a quantity
	// which would yield non-integer dimensional exponents,
    // or instantiate a quantity with all dimensions zero.

	template< bool B >
	struct static_assert;

	template<>
	struct static_assert< true >
	{
		typedef char dummy;
	};

	// The following batch of structs are type generators to calculate
	// the correct type of the result of various operations.

	template< typename LhsDims, typename RhsDims >
	struct product
	{
		enum { 
			d1 = LhsDims::dim1 + RhsDims::dim1,
			d2 = LhsDims::dim2 + RhsDims::dim2,
			d3 = LhsDims::dim3 + RhsDims::dim3,
			d4 = LhsDims::dim4 + RhsDims::dim4,
			d5 = LhsDims::dim5 + RhsDims::dim5,
			d6 = LhsDims::dim6 + RhsDims::dim6,
			d7 = LhsDims::dim7 + RhsDims::dim7 };
		typedef typename collapse< dimensions< d1, d2, d3, d4, d5, d6, d7 > >::type type;
	};

	template< typename LhsDims, typename RhsDims >
	struct quotient
	{
		enum {
			d1 = LhsDims::dim1 - RhsDims::dim1,
			d2 = LhsDims::dim2 - RhsDims::dim2,
			d3 = LhsDims::dim3 - RhsDims::dim3,
			d4 = LhsDims::dim4 - RhsDims::dim4,
			d5 = LhsDims::dim5 - RhsDims::dim5,
			d6 = LhsDims::dim6 - RhsDims::dim6,
			d7 = LhsDims::dim7 - RhsDims::dim7 };
		typedef typename collapse< dimensions< d1, d2, d3, d4, d5, d6, d7 > >::type type;
	};

	template< typename Dims >
	struct reciprocal
	{
		enum {
			d1 = - Dims::dim1,
			d2 = - Dims::dim2,
			d3 = - Dims::dim3,
			d4 = - Dims::dim4,
			d5 = - Dims::dim5,
			d6 = - Dims::dim6,
			d7 = - Dims::dim7 };
		typedef typename collapse< dimensions< d1, d2, d3, d4, d5, d6, d7 > >::type type;
	};

	template< typename Dims, int N >
	struct power
	{
		enum {
			d1 = N * Dims::dim1,
			d2 = N * Dims::dim2,
			d3 = N * Dims::dim3,
			d4 = N * Dims::dim4,
			d5 = N * Dims::dim5,
			d6 = N * Dims::dim6,
			d7 = N * Dims::dim7 };
		typedef typename collapse< dimensions< d1, d2, d3, d4, d5, d6, d7 > >::type type;
	};

	template< typename Dims, int N >
	struct root
	{
		enum { all_even_multiples =
			Dims::dim1 % N == 0 &&
			Dims::dim2 % N == 0 &&
			Dims::dim3 % N == 0 &&
			Dims::dim4 % N == 0 &&
			Dims::dim5 % N == 0 &&
			Dims::dim6 % N == 0 &&
			Dims::dim7 % N == 0 };
		enum {
			d1 = Dims::dim1 / N,
			d2 = Dims::dim2 / N,
			d3 = Dims::dim3 / N,
			d4 = Dims::dim4 / N,
			d5 = Dims::dim5 / N,
			d6 = Dims::dim6 / N,
			d7 = Dims::dim7 / N };
		typedef typename collapse< dimensions< d1, d2, d3, d4, d5, d6, d7 > >::type type;
	};

} // namespace detail

// "quantity" is the heart of the library.  It associates dimensions with a
// single "Rep" data member and protects it from dimensionally inconsistent use.

template< typename Dims >
class quantity
{
	public:

		quantity() {}
		// default copy constructor is ok
		// default assignment operator is ok
		// default destructor is ok

		// We need a "zero" of each type -- for comparisons, to initialize running
		// totals, etc.  Note:  0 m != 0 kg, since they are of different dimensionality.
		// The zero() function is really just defined for convenience, since
		// quantity< length_d >::zero() == 0 * meter(), etc.

		static quantity zero() { return quantity( Rep( 0.0 ) ); }

		// MSVC appears to get confused by certain binary operators and functions
		// which have two template parameters.  It has better luck when such things
		// are defined as members of the lhs.

		template< typename RhsDims >
			typename detail::product< Dims, RhsDims >::type
				operator*( const quantity< RhsDims > & rhs ) const
				{
					return detail::product< Dims, RhsDims >::type
						( detail::permit( m_value * rhs.get( detail::permit() ) ) );
				}

		template< typename RhsDims >
			typename detail::quotient< Dims, RhsDims >::type
				operator/( const quantity< RhsDims > & rhs ) const
				{
					return detail::quotient< Dims, RhsDims >::type
						( detail::permit( m_value / rhs.get( detail::permit() ) ) );
				}

		// The "permit" constructor and accessors allow access by functions that
		// should really just be declared as friends and refer directly to the
		// private members.

		explicit quantity( const detail::permit & p ) : m_value( p.m_value ) {}
		Rep & get( const detail::permit & ) { return m_value; }
		const Rep & get( const detail::permit & ) const { return m_value; }

	private:
		explicit quantity( Rep val ): m_value( val ) {}
		Rep m_value;

        enum{ ok = Dims::not_all_zero };
        typedef typename detail::static_assert< ok >::dummy 
            quantity_dimensions_must_not_all_be_zero;
};

// Give names to the seven fundamental dimensions of physical reality.

typedef dimensions< 1, 0, 0, 0, 0, 0, 0 > length_d;
typedef dimensions< 0, 1, 0, 0, 0, 0, 0 > mass_d;
typedef dimensions< 0, 0, 1, 0, 0, 0, 0 > time_interval_d;
typedef dimensions< 0, 0, 0, 1, 0, 0, 0 > electric_current_d;
typedef dimensions< 0, 0, 0, 0, 1, 0, 0 > thermodynamic_temperature_d;
typedef dimensions< 0, 0, 0, 0, 0, 1, 0 > amount_of_substance_d;
typedef dimensions< 0, 0, 0, 0, 0, 0, 1 > luminous_intensity_d;

// We define all the operators here.  Most could be defined as members of the
// quantity class, but that would really clutter up the class definition.
// It seems simpler to keep all the additions in one place, and so on.

// Addition operators

// quan += quan
template< typename Dims >
	inline quantity< Dims > & operator+=( quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		lhs.get( detail::permit() ) += rhs.get( detail::permit() );
		return lhs;
	}

// + quan
template< typename Dims >
	inline quantity< Dims > operator+( const quantity< Dims > & rhs )
	{
		return quantity< Dims >( detail::permit( + rhs.get( detail::permit() ) ) );
	}

// quan + quan
template< typename Dims >
	inline quantity< Dims > operator+( const quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		return quantity< Dims >( detail::permit( lhs.get( detail::permit() ) + rhs.get( detail::permit() ) ) );
	}

// Subtraction operators

// quan -= quan
template< typename Dims >
	inline quantity< Dims > & operator-=( quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		lhs.get( detail::permit() ) -= rhs.get( detail::permit() );
		return lhs;
	}

// - quan
template< typename Dims >
	inline quantity< Dims > operator-( const quantity< Dims > & rhs )
	{
		return quantity< Dims >( detail::permit( - rhs.get( detail::permit() ) ) );
	}

// quan - quan
template< typename Dims >
	inline quantity< Dims > operator-( const quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		return quantity< Dims >( detail::permit( lhs.get( detail::permit() ) - rhs.get( detail::permit() ) ) );
	}

// Multiplication operators

// quan *= num
template< typename Dims >
	inline quantity< Dims > operator*=( quantity< Dims > & lhs, const Rep & rhs )
	{
		lhs.get( detail::permit() ) *= rhs;
		return lhs;
	}

// quan * num
template< typename Dims >
	inline quantity< Dims > operator*( const quantity< Dims > & lhs, const Rep & rhs )
	{
		return quantity< Dims >( detail::permit( lhs.get( detail::permit() ) * rhs ) );
	}

// num * quan
template< typename Dims >
	inline quantity< Dims >operator*( const Rep & lhs, const quantity< Dims > & rhs )
	{
		return quantity< Dims >( detail::permit( lhs * rhs.get( detail::permit() ) ) );
	}

// quan * quan
// defined as member for portability

// Division operators

// quan /= num
template< typename Dims >
	inline quantity< Dims > operator/=( quantity< Dims > & lhs, const Rep & rhs )
	{
		lhs.get( detail::permit() ) /= rhs;
		return lhs;
	}

// quan / num
template< typename Dims >
	inline quantity< Dims > operator/( const quantity< Dims > & lhs, const Rep & rhs )
	{
		return quantity< Dims >( detail::permit( lhs.get( detail::permit() ) / rhs ) );
	}

// num / quan
template< typename Dims >
inline typename detail::reciprocal< Dims >::type operator/( const Rep & lhs, const quantity< Dims > & rhs )
	{
		return detail::reciprocal< Dims >::type
			( detail::permit( lhs / rhs.get( detail::permit() ) ) );
	}

// quan / quan
// defined as member for portability

// General powers

template< int N, typename Dims >
	typename detail::power< Dims, N >::type
		nth_power( const quantity< Dims > & rhs )
		{
			return detail::power< Dims, N >::type
				( detail::permit( Rep( pow( rhs.get( detail::permit() ), Rep( N ) ) ) ) );
		}

// Low powers defined separately for efficiency.

template< typename Dims >
	inline typename detail::power< Dims, 2 >::type
		square( const quantity< Dims > & arg )
		{
			return arg * arg;
		}

template< typename Dims >
	inline typename detail::power< Dims, 3 >::type
		cube( const quantity< Dims > & arg )
		{
			return arg * arg * arg;
		}

// General root

template< int N, typename Dims >
	typename detail::root< Dims, N >::type
		nth_root( const quantity< Dims > & rhs )
		{
			typedef typename detail::static_assert< detail::root< Dims, N >::all_even_multiples >::dummy
                root_result_dimensions_must_be_integral;
			return detail::root< Dims, N >::type
				( detail::permit( Rep( pow( rhs.get( detail::permit() ), Rep( 1.0 ) / N ) ) ) );
		}

// Low roots defined separately for convenience.

template< typename Dims >
	typename detail::root< Dims, 2 >::type 
		sqrt( const quantity< Dims > & rhs )
		{
			typedef typename detail::static_assert< detail::root< Dims, 2 >::all_even_multiples >::dummy
                root_result_dimensions_must_be_integral;
			return detail::root< Dims, 2 >::type
				( detail::permit( Rep( pow( rhs.get( detail::permit() ), Rep( 1.0 ) / 2 ) ) ) );
		}

// Comparison operators

template< typename Dims >
	inline bool operator==( const quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		return lhs.get( detail::permit() ) == rhs.get( detail::permit() );
	}

template< typename Dims >
	inline bool operator!=( const quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		return lhs.get( detail::permit() ) != rhs.get( detail::permit() );
	}

template< typename Dims >
	inline bool operator<=( const quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		return lhs.get( detail::permit() ) <= rhs.get( detail::permit() );
	}
template< typename Dims >
	inline bool operator>=( const quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		return lhs.get( detail::permit() ) >= rhs.get( detail::permit() );
	}

template< typename Dims >
	inline bool operator<( const quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		return lhs.get( detail::permit() ) < rhs.get( detail::permit() );
	}

template< typename Dims >
	inline bool operator>( const quantity< Dims > & lhs, const quantity< Dims > & rhs )
	{
		return lhs.get( detail::permit() ) > rhs.get( detail::permit() );
	}

// The seven SI base units.  These tie our numbers to the real world.

inline quantity< length_d > meter()
{
	return quantity< length_d >( detail::permit( Rep( 1.0 ) ) );
}

inline quantity< mass_d > kilogram()
{
	return quantity< mass_d >( detail::permit( Rep( 1.0 ) ) );
}

inline quantity< time_interval_d > second()
{
	return quantity< time_interval_d >( detail::permit( Rep( 1.0 ) ) );
}

inline quantity< electric_current_d > ampere()
{
	return quantity< electric_current_d >( detail::permit( Rep( 1.0 ) ) );
}

inline quantity< thermodynamic_temperature_d > kelvin()
{
	return quantity< thermodynamic_temperature_d >( detail::permit( Rep( 1.0 ) ) );
}

inline quantity< amount_of_substance_d > mole()
{
	return quantity< amount_of_substance_d >	( detail::permit( Rep( 1.0 ) ) );
}

inline quantity< luminous_intensity_d > candela()
{
	return quantity< luminous_intensity_d >	( detail::permit( Rep( 1.0 ) ) );
}

// The standard SI prefixes.

inline Rep yotta()	{ return Rep( 1e+24L ); }
inline Rep zetta()	{ return Rep( 1e+21L ); }
inline Rep exa()	{ return Rep( 1e+18L ); }
inline Rep peta()	{ return Rep( 1e+15L ); }
inline Rep tera()	{ return Rep( 1e+12L ); }
inline Rep giga()	{ return Rep( 1e+9L ); }
inline Rep mega()	{ return Rep( 1e+6L ); }
inline Rep kilo()	{ return Rep( 1e+3L ); }
inline Rep hecto()	{ return Rep( 1e+2L ); }
inline Rep deka()	{ return Rep( 1e+1L ); }
inline Rep deci()	{ return Rep( 1e-1L ); }
inline Rep centi()	{ return Rep( 1e-2L ); }
inline Rep milli()	{ return Rep( 1e-3L ); }
inline Rep micro()	{ return Rep( 1e-6L ); }
inline Rep nano()	{ return Rep( 1e-9L ); }
inline Rep pico()	{ return Rep( 1e-12L ); }
inline Rep femto()	{ return Rep( 1e-15L ); }
inline Rep atto()	{ return Rep( 1e-18L ); }
inline Rep zepto()	{ return Rep( 1e-21L ); }
inline Rep yocto()	{ return Rep( 1e-24L ); }

// Binary prefixes, pending adoption.

inline Rep kibi()	{ return Rep( 1024 ); }
inline Rep mebi()	{ return Rep( 1024 * kibi() ); }
inline Rep gibi()	{ return Rep( 1024 * mebi() ); }
inline Rep tebi()	{ return Rep( 1024 * gibi() ); }
inline Rep pebi()	{ return Rep( 1024 * tebi() ); }
inline Rep exbi()	{ return Rep( 1024 * pebi() ); }
inline Rep zebi()	{ return Rep( 1024 * exbi() ); }
inline Rep yobi()	{ return Rep( 1024 * zebi() ); }

// The rest of the standard dimensional types, as specified in SP811.

typedef dimensions< 2, 0, -2 >				absorbed_dose_d;
typedef dimensions< 2, 0, -3 >				absorbed_dose_rate_d;
typedef dimensions< 1, 0, -2 >				acceleration_d;
typedef dimensions< 0, 0, -1 >				activity_of_a_nuclide_d;
typedef dimensions< 0, 0, -1 >				angular_velocity_d;
typedef dimensions< 0, 0, -2 >				angular_acceleration_d;
typedef dimensions< 2, 0, 0 >				area_d;
typedef dimensions< -2, -1, 4, 2 >			capacitance_d;
typedef dimensions< -3, 0, 0, 0, 0, 1 >		concentration_d;
typedef dimensions< -2, 0, 0, 1 >			current_density_d;
typedef dimensions< 2, 0, -2 >				dose_equivalent_d;
typedef dimensions< -1, 1, -1 >				dynamic_viscosity_d;
typedef dimensions< 0, 0, 1, 1 >			electric_charge_d;
typedef dimensions< -3, 0, 1, 1 >			electric_charge_density_d;
typedef dimensions< -2, -1, 3, 2 >			electric_conductance_d;
typedef dimensions< 1, 1, -3, -1 >			electric_field_strenth_d;
typedef dimensions< -2, 0, 1, 1 >			electric_flux_density_d;
typedef dimensions< 2, 1, -3, -1 >			electric_potential_d;
typedef dimensions< 2, 1, -3, -2 >			electric_resistance_d;
typedef dimensions< 2, 1, -2 >				energy_d;
typedef dimensions< -1, 1, -2 >				energy_density_d;
typedef dimensions< 0, -1, 1, 1 >			exposure_d;
typedef dimensions< 1, 1, -2 >				force_d;
typedef dimensions< 0, 0, -1 >				frequency_d;
typedef dimensions< 2, 1, -2, 0, -1 >		heat_capacity_d;
typedef dimensions< 0, 1, -2 >				heat_density_d;
typedef dimensions< 0, 1, -3 >				heat_density_flow_rate_d;
typedef dimensions< 2, 1, -3 >				heat_flow_rate_d;
typedef dimensions< 0, 1, -3 >				heat_flux_density_d;
typedef dimensions< 0, 1, -3, 0, -1 >		heat_transfer_coefficient_d;
typedef dimensions< -2, 0, 0, 0, 0, 0, 1 >	illuminance_d;
typedef dimensions< 2, 1, -2, -2 >			inductance_d;
typedef dimensions< 0, 1, -3 >				irradiance_d;
typedef dimensions< 2, 0, -1 >				kinematic_viscosity_d;
typedef dimensions< -2, 0, 0, 0, 0, 0, 1 >	luminance_d;
typedef dimensions< 0, 0, 0, 0, 0, 0, 1 >	luminous_flux_d;
typedef dimensions< -1, 0, 0, 1 >			magnetic_field_strength_d;
typedef dimensions< 2, 1, -2, -1 >			magnetic_flux_d;
typedef dimensions< 0, 1, -2, -1 >			magnetic_flux_density_d;
typedef dimensions< 1, 1, -2, -2 >			magnetic_permeability_d;
typedef dimensions< -3, 1, 0 >				mass_density_d;
typedef dimensions< 0, 1, -1 >				mass_flow_rate_d;
typedef dimensions< 2, 1, -2, 0, 0, -1 >	molar_energy_d;
typedef dimensions< 2, 1, -2, -1, 0, -1 >	molar_entropy_d;
typedef dimensions< 2, 1, -2 >				moment_of_force_d;
typedef dimensions< -3, -1, 4, 2 >			permittivity_d;
typedef dimensions< 2, 1, -3 >				power_d;
typedef dimensions< -1, 1, -2 >				pressure_d;
typedef dimensions< 0, 1, -3 >				radiance_d;
typedef dimensions< 2, 1, -3 >				radiant_intensity_d;
typedef dimensions< 3, -1, 0 >				specific_volume_d;
typedef dimensions< 1, 0, -1 >				speed_d;
typedef dimensions< 2, 0, -2 >				specific_energy_d;
typedef dimensions< 2, 0, -2, 0, -1 >		specific_heat_capacity_d;
typedef dimensions< 3, -1, 0 >				specific_volume_d;
typedef dimensions< -1, 0, 1 >				substance_permeability_d;
typedef dimensions< 0, 1, -2 >				surface_tension_d;
typedef dimensions< 1, 1, -3, 0, -1 >		thermal_conductivity_d;
typedef dimensions< 2, 0, -1 >				thermal_diffusivity_d;
typedef dimensions< 0, -1, 3, 0, 1 >		thermal_insulance_d;
typedef dimensions< -2, -1, 3, 0, 1 >		thermal_resistance_d;
typedef dimensions< -1, -1, 3, 0, 1 >		thermal_resistivity_d;
typedef dimensions< 2, 1, -2 >				torque_d;
typedef dimensions< 3, 0, 0 >				volume_d;
typedef dimensions< 3, 0, -1 >				volume_flow_rate_d;
typedef dimensions< -1, 0, 0 >				wave_number_d;

// Handy values.

inline Rep									pi()			{ return Rep( 3.141592653589793238462L ); }
inline Rep									percent()		{ return Rep( 1 ) / 100; }

// Not approved for use alone, but needed for use with prefixes.

inline quantity< mass_d >					gram()			{ return kilogram() / 1000; }

// The derived SI units, as specified in SP811.

inline Rep									radian()		{ return Rep( 1 ); }
inline Rep									steradian()		{ return Rep( 1 ); }
inline quantity< force_d >					newton()		{ return meter() * kilogram() / square( second() ); }
inline quantity< pressure_d >				pascal()		{ return newton() / square( meter() ); }
inline quantity< energy_d >					joule()			{ return newton() * meter(); }
inline quantity< power_d >					watt()			{ return joule() / second(); }
inline quantity< electric_charge_d >		coulomb()		{ return second() * ampere(); }
inline quantity< electric_potential_d >		volt()			{ return watt() / ampere(); }
inline quantity< capacitance_d >			farad()			{ return coulomb() / volt(); }
inline quantity< electric_resistance_d >	ohm()			{ return volt() / ampere(); }
inline quantity< electric_conductance_d >	siemens()		{ return ampere() / volt(); }
inline quantity< magnetic_flux_d >			weber()			{ return volt() * second(); }
inline quantity< magnetic_flux_density_d >	tesla()			{ return weber() / square( meter() ); }
inline quantity< inductance_d >				henry()			{ return weber() / ampere(); }
inline quantity< thermodynamic_temperature_d > degree_celsius()	{ return kelvin(); }
inline quantity< luminous_flux_d >			lumen()			{ return candela() * steradian(); }
inline quantity< illuminance_d >			lux()			{ return lumen() / meter() / meter(); }
inline quantity< activity_of_a_nuclide_d >	becquerel()		{ return 1 / second(); }
inline quantity< absorbed_dose_d >			gray()			{ return joule() / kilogram(); }
inline quantity< dose_equivalent_d >		sievert()		{ return joule() / kilogram(); }
inline quantity< frequency_d >				hertz()			{ return 1 / second(); }

// The rest of the units approved for use with SI, as specified in SP811.
// (However, use of these units is generally discouraged.)

inline quantity< length_d >					angstrom()		{ return Rep( 1e-10L ) * meter(); }
inline quantity< area_d >					are()			{ return Rep( 1e+2L ) * square( meter() ); }
inline quantity< pressure_d >				bar()			{ return Rep( 1e+5L ) * pascal(); }
inline quantity< area_d >					barn()			{ return Rep( 1e-28L ) * square( meter() ); }
inline quantity< activity_of_a_nuclide_d >	curie()			{ return Rep( 3.7e+10L ) * becquerel(); }
inline quantity< time_interval_d >			day()			{ return Rep( 86400L ) * second(); }
inline Rep									degree_angle()  { return pi() / 180; }
inline quantity< acceleration_d >			gal()			{ return Rep( 1e-2L ) * meter() / square( second() ); }
inline quantity< area_d >					hectare()		{ return Rep( 1e+4L ) * square( meter() ); }
inline quantity< time_interval_d >			hour()			{ return Rep( 3600 ) * second(); }
inline quantity< speed_d >					knot()			{ return Rep( 1852 ) / 3600 * meter() / second(); }
inline quantity< volume_d >					liter()			{ return Rep( 1e-3L ) * cube( meter() ); }
inline quantity< time_interval_d >			minute()		{ return Rep( 60 ) * second(); }
inline Rep									minute_angle()	{ return pi() / 10800; }
inline quantity< length_d >					mile_nautical()	{ return Rep( 1852 ) * meter(); }
inline quantity< absorbed_dose_d >			rad()			{ return Rep( 1e-2L ) * gray(); }
inline quantity< dose_equivalent_d >		rem()			{ return Rep( 1e-2L ) * sievert(); }
inline quantity< exposure_d >				roentgen()		{ return Rep( 2.58e-4L ) * coulomb() / kilogram(); }
inline Rep									second_angle()	{ return pi() / 648000L; }
inline quantity< mass_d >					ton_metric()	{ return Rep( 1e+3L ) * kilogram(); }

// Alternate (non-US) spellings:

inline quantity< length_d >                 metre()         { return meter(); }
inline quantity< volume_d >                 litre()         { return liter(); }
inline Rep                                  deca()          { return deka(); }
inline quantity< mass_d >					tonne()			{ return ton_metric(); }

} // namespace quantity
} // namespace cuj

#endif // CUJ_QUANTITY_QUANTITY_HPP_MSK20010704
