/*
   Neutron Diffusion:
   Illustrate use of POOMA Arrays, stencil, multiple processor implementation.

   This program simulates the neutron fission in a critical mass of
   uranium 235.  The model and constants are based on five lectures
   given to incoming Los Alamos researchers in 1943, subsequently
   called "The Los Alamos Primer."  For details, see the book of the
   same title by Robert Serber, ISBN 0-520-07576-5.  The program
   models a two-dimensional, square mass of uranium 235.

   The program's command-line input are:
   1. the positive integral number 'p' of processors (for one dimension)
   2. the positive integral number 'n' of grid points (for one dimension)
   3. the desired number of time steps (a positive integer)
   4. the diffusion constant alpha (a double)
   Input parameters are for one dimension of this two-dimensional
   problem.  For example, a total of p*p processors are used.
*/

#include <cstdlib>
#include <cerrno>
#include "Pooma/Arrays.h"

// Define the stencil class performing one step of the neutron diffusion.
class Fisher
{
public:
  // Initialize the neutron diffusion stencil.
  // The alpha coefficient indicates the diffusion from a neighbor.
  // The beta coefficient indicates the net neutron creation from each fission.
  Fisher(const double alpha, const double beta)
    : alpha_(alpha), beta_(beta)
  {}

  // This stencil operator is applied to each interior domain position
  // (i,j).  The "C" template parameter permits use of this stencil
  // operator with both POOMA Arrays and Fields.
  template <class C>
  inline
  typename C::Element_t
  operator()(const C& c, int i, int j) const
  {
    return (                 alpha_ * c.read(i+1,j  ) +
			     alpha_ * c.read(i  ,j+1) +
	    (1.0 + beta_ - 4*alpha_)* c.read(i  ,j  ) +
			     alpha_ * c.read(i  ,j-1) +
			     alpha_ * c.read(i-1,j  ) );
  }

  inline int lowerExtent(int) const { return 1; }
  inline int upperExtent(int) const { return 1; }

private:

  const double alpha_;
  const double beta_;
};


// Process the command-line arguments, printing a message and 
// returning false if an error occurs.
bool processCommandLine(int argc, char*argv[],
			Inform& output,
			int& nuProcessors,
			int& nuGridPoints,
			int& nuDiffusions,
			double& alpha)
{
  // Check for the correct number of command-line arguments.
  if (argc != 5) {
    // Incorrect number of command-line arguments.
    output << argv[0]
	   << ": number-of-processors number-of-grid-points number-of-averagings alpha"
	   << std::endl;
    return false;
  }

  char *tail;		// Used to check for incorrect input.
  errno = 0;

  // Determine the number of processors for one of the two dimensions.
  nuProcessors = strtol(argv[1], &tail, 0);
  if (*tail != '\0' || errno == ERANGE || nuProcessors == 0) {
    output <<
      "The number of processors for one dimension must be a positive integer."
	   << std::endl;
    return false;
  }

  // Determine the number n of values along one dimension of
  // the grid.
  nuGridPoints = strtol(argv[2], &tail, 0);
  if (*tail != '\0' || errno == ERANGE || nuGridPoints == 0) {
    output <<
      "The number of grid points for one dimension must be a positive integer."
	   << std::endl;
    return false;
  }
  // The dimension must be a multiple of the number of processors
  // since we are using a UniformGridLayout.
  nuGridPoints = ((nuGridPoints+nuProcessors-1) / nuProcessors) * nuProcessors;

  // Determine the number of averagings.
  nuDiffusions = strtol(argv[3], &tail, 0);
  if (*tail != '\0'|| errno == ERANGE || nuDiffusions == 0) {
    output << "The number of diffusions be a positive integer."
	   << std::endl;
    return false;
  }

  // Determine the diffusion constant.
  alpha = strtod(argv[4], &tail);
  if (*tail != '\0' || errno == ERANGE ||
      !(0.0 < alpha && alpha < 0.5)) {
    output << "The diffusion constant alpha must be in the range (0.0, 0.5)."
	   << std::endl;
    return false;
  }

  // Successful parsing.
  return true;
}


int main(int argc, char* argv[])
{
  // Prepare the POOMA library for execution.
  Pooma::initialize(argc,argv);

  // Since multiple copies of this program may simultaneously run, we
  // cannot use standard input and output.  Instead we use command-line
  // arguments, which are replicated, for input, and we use an Inform
  // stream for output.
  Inform output;

  // The number of processors to use along one dimension.
  int nuProcessors;
  // The number n of values along one dimension of the grid.
  int n;
  // Determine the number of diffusions and iterations.
  // Each iteration performs two diffusions.
  int nuDiffusions;

  // The diffusion constant.
  double alpha;

  // Read the program input from the command-line arguments.
  if (!processCommandLine(argc, argv, output,
			  nuProcessors, n, nuDiffusions, alpha))
    return EXIT_FAILURE;

  // The net neutron creation rate for each fission.
  const double beta = 2.0 /* 2 created */ - 1.0 /* one consumed */;

  // Specify the arrays' domains [0,n) x [0,n).
  Interval<1> N(0, n-1);
  Interval<2> vertDomain(N, N);

  // Set up interior domains [1,n-1) x [1,n-1) for computation.
  Interval<1> I(1,n-2);
  Interval<2> interiorDomain(I,I);

  // Create the distributed arrays.
  // Partition the arrays' domains uniformly, i.e., each patch has the
  // same size.  The first parameter tells how many patches for each
  // dimension.  Guard layers optimize communication between patches.
  // Internal guards surround each patch.  External guards surround
  // the entire array domain.
  UniformGridPartition<2> partition(Loc<2>(nuProcessors, nuProcessors),
				    GuardLayers<2>(1),  // internal
				    GuardLayers<2>(0)); // external
  UniformGridLayout<2> layout(vertDomain, partition, DistributedTag());

  // The Array template parameters indicate 2 dimensions and a 'double'
  // value type.  MultiPatch indicates multiple computation patches,
  // i.e., distributed computation.  The UniformTag indicates the
  // patches should have the same size.  Each patch has Brick type.
  Array<2, double, MultiPatch<UniformTag, Remote<Brick> > > Na(layout);
  Array<2, double, MultiPatch<UniformTag, Remote<Brick> > > Nb(layout);

  // Set up the initial conditions.
  // All grid values should be zero except for the central value.
  Na = Nb = 0.0;
  // Ensure all data-parallel computation finishes before accessing a value.
  Pooma::blockAndEvaluate();

  // Start the fission with a neutron source at the center.
  Nb(n/2,n/2) = 1.0e+7;

  // Create the stencil performing the computation.
  Fisher fisher(alpha, beta);
  Stencil<Fisher> stencil(fisher);

  // Perform the simulation.
  for (int k = 0; k < nuDiffusions; k += 2) {
    // Read from Nb.  Write to Na.
    Na(interiorDomain) = stencil(Nb, interiorDomain);

    // Read from Na.  Write to Nb.
    Nb(interiorDomain) = stencil(Na, interiorDomain);
  }

  // Print out the final central value.
  Pooma::blockAndEvaluate();	// Ensure all computation has finished.
  output << (nuDiffusions % 2 ? Na(n/2,n/2) : Nb(n/2,n/2)) << std::endl;

  // The arrays are automatically deallocated.

  // Tell the POOMA library execution has finished.
  Pooma::finalize();
  return EXIT_SUCCESS;
}
